import os
import sys
from typing import Any


class MetaInfo:
    """元信息类

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(self) -> None:
        #: melobot 版本
        self.VER: str = "2.5.8"
        #: melobot 项目名称
        self.PROJ_NAME: str = "melobot"
        #: melobot 项目描述
        self.PROJ_DESC: str = (
            "A qq bot development framework with friendly APIs, session control and plugin-supported."
        )
        #: melobot 项目地址
        self.PROJ_SRC: str = "https://github.com/aicorein/melobot"
        #: melobot 项目作者
        self.AUTHOR: str = "aicorein"
        #: melobot 项目作者邮箱
        self.AUTHOR_EMAIL: str = "melodyecho@glowmem.com"
        #: 当前运行的 argv
        self.ARGV: list[str] = sys.argv
        #: 当前系统平台
        self.PLATFORM: str = sys.platform
        #: 当前 python 版本
        self.PY_VER: str = sys.version
        #: 当前 python 信息
        self.PY_INFO: sys._version_info = sys.version_info
        #: 当前系统路径分隔符号，如 win 平台下的 "\\"
        self.OS_SEP: str = os.sep
        #: 当前系统路径间的分隔符号，如 win 平台下的 ";"
        self.PATH_SEP: str = os.pathsep
        #: 当前系统行尾序列，如 win 平台下的 "\\r\\n"
        self.LINE_SEP: str = os.linesep
        #: 当前运行的环境变量
        self.ENV: os._Environ[str] = os.environ

    def get_all(self) -> dict[str, Any]:
        """以字典形式获取所有元信息

        :return: 包含所有元信息的，属性名为键的字典
        """
        return dict(self.__dict__.items())


MELOBOT_LOGO = r"""  __  __      _       ____        _
 |  \/  | ___| | ___ | __ )  ___ | |_
 | |\/| |/ _ \ |/ _ \|  _ \ / _ \| __|
 | |  | |  __/ | (_) | |_) | (_) | |_
 |_|  |_|\___|_|\___/|____/ \___/ \__|
"""
MELOBOT_LOGO_LEN = max(len(_) for _ in MELOBOT_LOGO.split("\n"))



import sys
from enum import Enum
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Coroutine,
    Literal,
    NamedTuple,
    Optional,
    OrderedDict,
    ParamSpec,
    Type,
    TypeAlias,
    TypedDict,
    TypeVar,
    Union,
    cast,
)

import better_exceptions
from typing_extensions import NotRequired

# 修复在 windows powershell 显示错误的 bug
better_exceptions.encoding.ENCODING = sys.stdout.encoding
better_exceptions.formatter.ENCODING = sys.stdout.encoding
better_exceptions.hook()


class CQMsgDict(TypedDict):
    """onebot 标准的消息段对象"""

    type: str
    data: dict[str, float | int | str]


class CustomNodeData(TypedDict):
    """onebot 标准的自定义消息结点数据"""

    name: str
    uin: str
    content: list[CQMsgDict]
    seq: NotRequired[list[CQMsgDict]]


class ReferNodeData(TypedDict):
    """onebot 标准的引用消息结点数据"""

    id: str


class MsgNodeDict(TypedDict):
    """onebot 标准的转发消息结点"""

    type: Literal["node"]
    data: CustomNodeData | ReferNodeData


class ParseArgs:
    # 命令参数类

    def __init__(self, values: list[Any] | None) -> None:
        self.vals = values
        self.formatted = False


class LogicMode(Enum):
    """逻辑模式枚举类型"""

    AND = 1
    OR = 2
    NOT = 3
    XOR = 4

    @classmethod
    def calc(cls, logic: "LogicMode", v1: Any, v2: Any = None) -> bool:
        if logic == LogicMode.AND:
            return (v1 and v2) if v2 is not None else bool(v1)
        elif logic == LogicMode.OR:
            return (v1 or v2) if v2 is not None else bool(v1)
        elif logic == LogicMode.NOT:
            return not v1
        elif logic == LogicMode.XOR:
            return (v1 ^ v2) if v2 is not None else bool(v1)

    @classmethod
    def seq_calc(cls, logic: "LogicMode", values: list[Any]) -> bool:
        if len(values) <= 0:
            return False
        elif len(values) <= 1:
            return bool(values[0])

        idx = 0
        res: bool
        while idx < len(values):
            if idx == 0:
                res = cls.calc(logic, values[idx], values[idx + 1])
                idx += 1
            else:
                res = cls.calc(logic, res, values[idx])
            idx += 1
        return res


class User(int, Enum):
    """用户权限等级枚举"""

    OWNER = 10000
    SU = 1000
    WHITE = 100
    USER = 10
    BLACK = -1


class PriorLevel(int, Enum):
    """事件处理器优先级枚举

    为方便进行优先级设置，有 MIN, MAX, MEAN 三个枚举值
    """

    MIN = 0
    MAX = 1000
    MEAN = (MAX + MIN) // 2


class BotLife(Enum):
    """bot 实例的生命周期枚举"""

    LOADED = 1
    CONNECTED = 2
    BEFORE_CLOSE = 3
    BEFORE_STOP = 4
    EVENT_BUILT = 5
    ACTION_PRESEND = 6


#: 泛型 T，无约束
T = TypeVar("T")
#: 泛型 T，无约束
T1 = TypeVar("T1")
#: 泛型 T，无约束
T2 = TypeVar("T2")
#: 泛型 T，无约束
T3 = TypeVar("T3")
#: :obj:`typing.ParamSpec` 泛型 P，无约束
P = ParamSpec("P")


class Void:
    """“无值”对象，而不是 :obj:`None` 代表的“空值”

    使用方法：直接使用这个类即可。

    这个对象的类型标注是：:obj:`VoidType`。
    """

    pass


#: “无值”对象类型
VoidType: TypeAlias = Type[Void]



import asyncio
from contextvars import ContextVar, Token
from copy import deepcopy
from functools import wraps

from ..base.abc import SessionRule
from ..base.exceptions import BotSessionError, SessionHupTimeout
from ..base.tools import get_twin_event, to_task
from ..base.typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Optional,
    P,
    Type,
    Union,
    Void,
    VoidType,
)

if TYPE_CHECKING:
    from ..base.abc import BotAction, BotEvent
    from ..bot.init import BotLocal, MeloBot
    from ..models.event import (
        MessageEvent,
        MetaEvent,
        NoticeEvent,
        RequestEvent,
        ResponseEvent,
    )
    from ..plugin.handler import EventHandler


class BotSession:
    """Bot Session 类。不需要直接实例化，必须通过 BotSessionBuilder 构造。"""

    def __init__(
        self,
        event: Union["MessageEvent", "RequestEvent", "MetaEvent", "NoticeEvent"],
        manager: Type["BotSessionManager"],
        space_tag: Optional["EventHandler"] = None,
    ) -> None:
        super().__init__()
        self.store: dict[str, Any] = {}
        # 永远指向当前上下文的 event
        self.event = event

        self._manager = manager
        # session 是否空闲的标志，由 BotSessionManager 修改和管理
        self._free_signal = asyncio.Event()
        self._free_signal.set()
        # session 是否挂起的标志。二者互为孪生反状态。由 BotSessionManager 修改和管理
        # 注意 session 挂起时一定是非空闲和非过期的
        self._hup_signal, self._awake_signal = get_twin_event()
        # session 是否过期的标志，由 BotSessionManager 修改和管理
        self._expired = False
        # 用于标记该 session 属于哪个 session 空间，如果为 None 则表明是一次性 session
        # 其实这里如果传入 space_tag 则一定是所属 handler 的引用
        self._space_tag: Optional["EventHandler"] = space_tag
        # 所属 handler 的引用（和 space_tag 不一样，所有在 handler 中产生的 session，此属性必为非空）
        self._handler: Optional["EventHandler"] = None

    @property
    def args(self):
        if (
            self._handler is not None
            and hasattr(self._handler, "parser")
            and self.event._args_map is not None
        ):
            args_group = self.event._args_map.get(self._handler.parser.id)
            if args_group is None:
                return None
            for cmd_name in self._handler.parser.target:
                args = args_group.get(cmd_name)
                if args is not None:
                    return deepcopy(args.vals)
            raise BotSessionError("尝试获取的命令解析参数不存在")
        return None

    async def hup(self, overtime: Optional[int] = None) -> None:
        """当前 session 挂起（也就是所在方法的挂起）。直到满足同一 session_rule 的事件重新进入， session
        所在方法便会被唤醒。但如果设置了超时时间且在唤醒前超时，则会强制唤醒 session， 并抛出 SessionHupTimeout 异常."""
        self._manager._hup(self)
        if overtime is None:
            await self._awake_signal.wait()
        elif overtime <= 0:
            raise BotSessionError("session 挂起超时时间（秒数）必须 > 0")
        else:
            await asyncio.wait(
                [to_task(self._awake_signal.wait()), to_task(asyncio.sleep(overtime))],
                return_when=asyncio.FIRST_COMPLETED,
            )
            if self._awake_signal.is_set():
                return
            self._manager._rouse(self)
            raise SessionHupTimeout("session 挂起超时")

    def destory(self) -> None:
        """销毁方法。 其他 session 调用会立即清空 session 存储、事件记录、挂起时间记录。 如果调用 session 有
        space_tag，还会从存储空间中移除该 session."""
        self._manager._expire(self)


class SessionLocal:
    """Session 自动上下文."""

    __slots__ = tuple(
        list(filter(lambda x: not (len(x) >= 2 and x[:2] == "__"), dir(BotSession)))
        + ["__storage__"]
    )

    def __init__(self) -> None:
        object.__setattr__(self, "__storage__", ContextVar("session_ctx"))
        self.__storage__: ContextVar[BotSession | VoidType]

    def _get_var(self) -> BotSession:
        var = self.__storage__.get()
        if var is Void:
            raise LookupError("上下文当前为 Void，识别为跨 bot 通信")
        return var  # type: ignore

    def __setattr__(self, __name: str, __value: Any) -> None:
        setattr(self._get_var(), __name, __value)

    def __getattr__(self, __name: str) -> Any:
        return getattr(self._get_var(), __name)

    def _add_ctx(self, ctx: BotSession | VoidType) -> Token:
        return self.__storage__.set(ctx)

    def _del_ctx(self, token: Token) -> None:
        self.__storage__.reset(token)


SESSION_LOCAL = SessionLocal()


class AttrSessionRule(SessionRule):
    """属性会话规则（通过事件属性判断两个事件是否在同一会话中）"""

    def __init__(self, *attrs: Any) -> None:
        """初始化一个属性会话规则

        :param attrs: 用于判断的属性链

        .. code:: python

           # 假设某类型事件拥有以下的属性链：
           event.a.b.c
           # 实例化一个属性会话规则：
           rule = AttrSessionRule('a', 'b', 'c')
           '''
           此后该规则可用于会话判断。当两事件满足：event1.a.b.c == event2.a.b.c 时，
           两事件即被认为在同一会话中
           '''

           # 试想，如果需要对消息事件做这样的会话控制：“同一个 qq 号发出的消息事件都在同一会话中”
           # 写一个这样的规则就可以了：
           rule = AttrSessionRule('sender', 'id')
           # 因为 MessageEvent 类型的事件的 sender 属性的 id 属性正好对应发送者 qq 号
        """
        super().__init__()
        self.attrs = attrs

    def _get_val(self, e: "BotEvent", attrs: tuple[str, ...]) -> Any:
        val = e
        try:
            for attr in attrs:
                val = getattr(val, attr)
        except AttributeError:
            raise BotSessionError(f"session 规则指定的属性 {attr} 不存在")
        return val

    def compare(self, e1: "BotEvent", e2: "BotEvent") -> bool:
        return self._get_val(e1, self.attrs) == self._get_val(e2, self.attrs)


class BotSessionManager:
    STORAGE: dict["EventHandler", set[BotSession]] = {}
    HUP_STORAGE: dict["EventHandler", set[BotSession]] = {}
    # 各个 handler 对饮的操作锁
    WORK_LOCKS: dict["EventHandler", asyncio.Lock] = {}
    # 用来标记 cls.get 等待一个挂起的 session 时的死锁
    DEADLOCK_FLAGS: dict["EventHandler", asyncio.Event] = {}
    # 对应每个 handler 的 try_attach 过程的操作锁
    ATTACH_LOCKS: dict["EventHandler", asyncio.Lock] = {}
    BOT: "MeloBot"  # type: ignore

    @classmethod
    def _bind(cls, bot_local_var: "BotLocal") -> None:
        cls.BOT = bot_local_var  # type: ignore

    @classmethod
    def register(cls, handler: "EventHandler") -> None:
        """以 handler 为键，注册 handler 对应的 session 空间、操作锁和挂起 session 空间."""
        cls.STORAGE[handler] = set()
        cls.WORK_LOCKS[handler] = asyncio.Lock()
        cls.HUP_STORAGE[handler] = set()
        cls.DEADLOCK_FLAGS[handler] = asyncio.Event()
        cls.ATTACH_LOCKS[handler] = asyncio.Lock()

    @classmethod
    def inject(cls, session: BotSession, handler: "EventHandler") -> None:
        """Handler 内绑定 handler 引用到 session."""
        session._handler = handler

    @classmethod
    def _attach(
        cls,
        event: Union["MessageEvent", "RequestEvent", "MetaEvent", "NoticeEvent"],
        handler: "EventHandler",
    ) -> bool:
        """Session 附着操作，临界区操作。只能在 cls.try_attach 中进行."""
        session = None
        for s in cls.HUP_STORAGE[handler]:
            # session 的挂起方法，保证 session 一定未过期，因此不进行过期检查
            if handler._rule.compare(s.event, event):  # type: ignore
                session = s
                break
        # 如果获得一个挂起的 session，它一定是可附着的，附着后需要唤醒
        if session:
            session.event = event
            cls._rouse(session)
            return True
        return False

    @classmethod
    async def try_attach(
        cls,
        event: Union["MessageEvent", "RequestEvent", "MetaEvent", "NoticeEvent"],
        handler: "EventHandler",
    ) -> bool:
        """检查是否有挂起的 session 可供 event 附着。 如果有则附着并唤醒，并返回 True。否则返回 False。"""
        if handler._rule is None:
            return False

        async with cls.ATTACH_LOCKS[handler]:
            t1 = asyncio.create_task(cls.DEADLOCK_FLAGS[handler].wait(), name="flag")
            t2 = asyncio.create_task(cls.WORK_LOCKS[handler].acquire(), name="lock")
            done, _ = await asyncio.wait([t1, t2], return_when=asyncio.FIRST_COMPLETED)
            # 等待完成后，一定要记得取消另一个任务！否则可能异常加锁
            if done.pop().get_name() == "flag":
                res = cls._attach(event, handler)
                cls.DEADLOCK_FLAGS[handler].clear()
                t2.cancel()
                return res
            else:
                res = cls._attach(event, handler)
                cls.WORK_LOCKS[handler].release()
                t1.cancel()
                return res

    @classmethod
    def _hup(cls, session: BotSession) -> None:
        """挂起 session."""
        if session._space_tag is None:
            raise BotSessionError(
                "当前 session 上下文因为缺乏 session_rule 作为唤醒标志，无法挂起"
            )
        elif session._expired:
            raise BotSessionError("过期的 session 不能被挂起")
        cls.STORAGE[session._space_tag].remove(session)
        cls.HUP_STORAGE[session._space_tag].add(session)
        session._awake_signal.clear()

    @classmethod
    def _rouse(cls, session: BotSession) -> None:
        """唤醒 session."""
        cls.HUP_STORAGE[session._space_tag].remove(session)  # type: ignore
        cls.STORAGE[session._space_tag].add(session)  # type: ignore
        session._awake_signal.set()

    @classmethod
    def _expire(cls, session: BotSession) -> None:
        """标记该 session 为过期状态，并进行销毁操作（如果存在于某个 session_space，则从中移除）"""
        if session._expired:
            return
        session.store.clear()
        session._expired = True
        if session._space_tag:
            cls.STORAGE[session._space_tag].remove(session)

    @classmethod
    def recycle(cls, session: BotSession, alive: bool = False) -> None:
        """Session 所在方法运行结束后，回收 session。 默认将 session 销毁。若 alive 为 True，则保留."""
        session._free_signal.set()
        if not alive:
            cls._expire(session)

    @classmethod
    async def get(
        cls,
        event: Union["MessageEvent", "RequestEvent", "MetaEvent", "NoticeEvent"],
        handler: "EventHandler",
    ) -> Optional[BotSession]:
        """Handler 内获取 session 方法。自动根据 handler._rule 判断是否需要映射到 session_space 进行存储。
        然后根据具体情况，获取已有 session 或新建 session。当尝试获取非空闲 session 时，如果 handler 指定不等待则返回
        None."""
        if handler._rule:
            # session_space, session._free_signal 竞争，需要加锁
            async with cls.WORK_LOCKS[handler]:
                session = await cls._get_on_rule(event, handler)
                # 必须在锁的保护下修改 session._free_signal
                if session:
                    session._free_signal.clear()
        else:
            session = cls._make(event, handler)
            session._free_signal.clear()

        return session

    @classmethod
    def _make(
        cls,
        event: Union["MessageEvent", "RequestEvent", "MetaEvent", "NoticeEvent"],
        handler: Optional["EventHandler"] = None,
    ) -> BotSession:
        """内部使用的创建 session 方法。如果 handler 为空，即缺乏 space_tag，则为一次性 session。 或 handler._rule
        为空，则也为一次性 session."""
        if handler:
            if handler._rule:
                session = BotSession(event, cls, handler)
                cls.STORAGE[handler].add(session)
                return session
        session = BotSession(event, cls)
        return session

    @classmethod
    def make_temp(
        cls, event: Union["MessageEvent", "RequestEvent", "MetaEvent", "NoticeEvent"]
    ) -> BotSession:
        """创建一次性 session。确定无需 session 管理机制时可以使用。 否则一定使用 cls.get 方法."""
        return cls._make(event)

    @classmethod
    async def _get_on_rule(
        cls,
        event: Union["MessageEvent", "RequestEvent", "MetaEvent", "NoticeEvent"],
        handler: "EventHandler",
    ) -> Optional[BotSession]:
        """根据 handler 具体情况，从对应 session_space 中获取 session 或新建 session。 或从
        hup_session_space 中唤醒 session，或返回 None."""
        session = None
        check_rule, session_space, hup_session_space, conflict_wait = (
            handler._rule,
            cls.STORAGE[handler],
            cls.HUP_STORAGE[handler],
            handler._wait_flag,
        )

        # for 循环都需要即时 break，保证遍历 session_space 时没有协程切换。因为切换后 session_space 可能发生变动
        for s in session_space:
            if check_rule.compare(s.event, event) and not s._expired:  # type: ignore
                session = s
                break
        # 如果会话不存在，生成一个新 session 变量
        if session is None:
            return cls._make(event, handler)
        # 如果会话存在，且未过期，且空闲，则附着到这个 session 上
        if session._free_signal.is_set():
            session.event = event
            return session
        # 如果会话存在，且未过期，但是不空闲，选择不等待
        if not conflict_wait:
            return None
        # 如果会话存在，且未过期，但是不空闲，选择等待，此时就不得不陷入等待（即将发生协程切换）
        await asyncio.wait(
            [to_task(session._free_signal.wait()), to_task(session._hup_signal.wait())],
            return_when=asyncio.FIRST_COMPLETED,
        )
        if session._hup_signal.is_set():
            cls.DEADLOCK_FLAGS[handler].set()
            await session._free_signal.wait()
        """
        重新切换回本协程后，session 有可能变为过期，但此时一定是空闲的。
        同时一定是非挂起状态。因为上面解决了可能存在的挂起死锁问题。
        即使该 session 因过期被所在的 session_space 清除也无妨，因为此处有引用，
        该 session 并不会消失。且此处不操作 session_space，无需担心 session_space 变动
        """
        # 如果过期，生成一个新的 session 变量
        if session._expired:
            return cls._make(event, handler)
        # 如果未过期，则附着到这个 session 上
        else:
            session.event = event
            return session

    @classmethod
    def _activate(
        cls, action_getter: Callable[P, Coroutine[Any, Any, "BotAction"]]
    ) -> Callable[P, Coroutine[Any, Any, Union["BotAction", "ResponseEvent", None]]]:
        """对 action 构造器进行装饰，使产生的 action “活化”。 让其能自动识别上下文，自动附加触发 event，并自动完成发送过程."""

        @wraps(action_getter)
        async def activated_action(
            *args: Any, **kwargs: Any
        ) -> Union["BotAction", "ResponseEvent", None]:
            try:
                if SESSION_LOCAL._expired:
                    raise BotSessionError(
                        "当前  session 上下文已有过期标记，无法再执行 action 操作"
                    )
            except LookupError:
                pass

            action: "BotAction" = await action_getter(*args, **kwargs)
            try:
                action._fill_trigger(SESSION_LOCAL.event)
            except LookupError:
                pass

            if not action._ready:
                return action
            if action.resp_id is None:
                await cls.BOT._responder.take_action(action)
                return None
            else:
                return await (await cls.BOT._responder.take_action_wait(action))

        return activated_action


def any_event() -> Union["MessageEvent", "RequestEvent", "MetaEvent", "NoticeEvent"]:
    """获取当前会话下的事件

    :return: 具体的事件
    """
    try:
        return SESSION_LOCAL.event
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")


def msg_event() -> "MessageEvent":
    """获取当前会话下的事件

    确定此事件是消息事件时使用，例如在消息事件的处理函数中，使用该方法即可获得精准的类型提示

    :return: 具体的事件
    """
    return any_event()  # type: ignore


def notice_event() -> "NoticeEvent":
    """获取当前会话下的事件

    确定此事件是通知事件时使用，例如在通知事件的处理函数中，使用该方法即可获得精准的类型提示

    :return: 具体的事件
    """
    return any_event()  # type: ignore


def req_event() -> "RequestEvent":
    """获取当前会话下的事件

    确定此事件是请求事件时使用，例如在请求事件的处理函数中，使用该方法即可获得精准的类型提示

    :return: 具体的事件
    """
    return any_event()  # type: ignore


def meta_event() -> "MetaEvent":
    """获取当前会话下的事件

    确定此事件是元事件时使用，例如在元事件的处理函数中，使用该方法即可获得精准的类型提示

    :return: 具体的事件
    """
    return any_event()  # type: ignore


def msg_text() -> str:
    """获取当前会话下的消息事件的，所有纯文本消息的合并字符串。
    等价于手动读取消息事件的 :attr:`~.MessageEvent.text` 属性

    只能在确定当前会话下必为消息事件时使用

    :return: 所有纯文本消息的合并字符串
    """
    event: "MessageEvent" = any_event()  # type: ignore
    return event.text


def msg_args() -> list[Any] | None:
    """获取当前会话下的消息事件的所有解析参数值。

    只能在确定当前会话下必为消息事件时使用

    :return: 解析参数值列表或 :obj:`None`
    """
    try:
        return SESSION_LOCAL.args
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")


def get_store() -> dict:
    """返回当前会话的存储字典对象，可直接操作

    会话存储字典对象用于在会话层面保存和共享数据。

    :return: 会话的存储字典对象
    """
    try:
        return SESSION_LOCAL.store
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")


async def pause(overtime: Optional[int] = None) -> None:
    """会话暂停直到被同一会话的事件唤醒

    暂时停止本会话。当本会话的会话规则判断有属于本会话的另一事件发生，
    本会话将自动被唤醒。

    可指定超时时间。如果超时时间结束会话仍未被唤醒，此时将会被强制唤醒，
    并抛出一个 :class:`.SessionHupTimeout` 异常。可以捕获此异常，
    实现自定义的超时处理逻辑

    :param overtime: 超时时间
    """
    try:
        await SESSION_LOCAL.hup(overtime)
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")


def dispose() -> None:
    """销毁当前会话

    将会清理会话的存储空间，并将会话标记为过期态。
    此时调用该方法的函数依然可以运行，但是此会话状态下无法再进行行为操作。

    一般来说，会话将会自动销毁。
    只有在注册事件处理函数时使用 `session_hold=True`，才会需要使用此函数。

    """
    try:
        SESSION_LOCAL.destory()
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")



import asyncio
import inspect
import io
import os
import pathlib
import time
from asyncio import iscoroutine, iscoroutinefunction
from contextlib import asynccontextmanager
from functools import wraps

import rich

from .exceptions import BotBaseUtilsError, BotRuntimeError
from .typing import T1, T2, T3, Any, Callable, Coroutine, Optional, P, T


class Singleton:
    def __new__(cls, *args: Any, **kwargs: Any):
        if not hasattr(cls, "__instance__"):
            cls.__instance__ = super(Singleton, cls).__new__(cls)
        return cls.__instance__


class AsyncTwinEvent(asyncio.Event):
    """孪生 Event，会和绑定的一方时刻保持状态相反。"""

    def __init__(self) -> None:
        super().__init__()
        self._twin: Optional[AsyncTwinEvent] = None

    def bind(self, twin: "AsyncTwinEvent") -> None:
        self._twin = twin
        if self.is_set():
            super(AsyncTwinEvent, self._twin).clear()
        else:
            super(AsyncTwinEvent, self._twin).set()

    def set(self) -> None:
        super().set()
        if self._twin:
            super(AsyncTwinEvent, self._twin).clear()

    def clear(self) -> None:
        super().clear()
        if self._twin:
            super(AsyncTwinEvent, self._twin).set()


def get_twin_event() -> tuple[asyncio.Event, asyncio.Event]:
    """获得两个时刻保持状态相反的 asyncio.Event。 获得的第一个为 unset，另一个为 set."""
    a, b = AsyncTwinEvent(), AsyncTwinEvent()
    a.bind(b)
    b.bind(a)
    return a, b


class RWController:
    """异步读写控制器

    提供异步安全的读写上下文。在读取时可以多读，同时读写互斥。

    使用方法：

    .. code:: python

       rwc = RWController()
       # 读时使用此控制器的安全读上下文：
       async with rwc.safe_read():
           ...
       # 写时使用此控制器的安全写上下文：
       async with rwc.safe_write():
           ...
    """

    def __init__(self, read_limit: Optional[int] = None) -> None:
        """初始化一个异步读写控制器

        :param read_limit: 读取的数量限制，为空则不限制
        """
        write_semaphore = asyncio.Semaphore(1)
        if read_limit:
            read_semaphore = asyncio.Semaphore(read_limit)
        else:
            read_semaphore = None
        read_num = 0
        read_num_lock = asyncio.Lock()

        @asynccontextmanager
        async def safe_read():
            nonlocal read_num, read_semaphore, write_semaphore, read_num_lock
            if read_semaphore:
                await read_semaphore.acquire()
            async with read_num_lock:
                if read_num == 0:
                    await write_semaphore.acquire()
                read_num += 1
            try:
                yield
            finally:
                async with read_num_lock:
                    read_num -= 1
                    if read_num == 0:
                        write_semaphore.release()
                    if read_semaphore:
                        read_semaphore.release()

        @asynccontextmanager
        async def safe_write():
            nonlocal write_semaphore
            await write_semaphore.acquire()
            try:
                yield
            finally:
                write_semaphore.release()

        self.safe_read = safe_read
        self.safe_write = safe_write


class IdWorker:
    """雪花算法生成 ID."""

    def __init__(self, datacenter_id, worker_id, sequence=0) -> None:
        self.MAX_WORKER_ID = -1 ^ (-1 << 3)
        self.MAX_DATACENTER_ID = -1 ^ (-1 << 5)
        self.WOKER_ID_SHIFT = 12
        self.DATACENTER_ID_SHIFT = 12 + 3
        self.TIMESTAMP_LEFT_SHIFT = 12 + 3 + 5
        self.SEQUENCE_MASK = -1 ^ (-1 << 12)
        self.STARTEPOCH = 1064980800000
        # sanity check
        if worker_id > self.MAX_WORKER_ID or worker_id < 0:
            raise ValueError("worker_id 值越界")
        if datacenter_id > self.MAX_DATACENTER_ID or datacenter_id < 0:
            raise ValueError("datacenter_id 值越界")
        self.worker_id = worker_id
        self.datacenter_id = datacenter_id
        self.sequence = sequence
        self.last_timestamp = -1

    def __gen_timestamp(self) -> int:
        """生成整数时间戳."""
        return int(time.time() * 1000)

    def get_id(self) -> int:
        """获取新 ID."""
        timestamp = self.__gen_timestamp()

        # 时钟回拨
        if timestamp < self.last_timestamp:
            raise ValueError(f"时钟回拨，{self.last_timestamp} 前拒绝 id 生成请求")
        if timestamp == self.last_timestamp:
            self.sequence = (self.sequence + 1) & self.SEQUENCE_MASK
            if self.sequence == 0:
                timestamp = self.__til_next_millis(self.last_timestamp)
        else:
            self.sequence = 0
        self.last_timestamp = timestamp
        new_id = (
            ((timestamp - self.STARTEPOCH) << self.TIMESTAMP_LEFT_SHIFT)
            | (self.datacenter_id << self.DATACENTER_ID_SHIFT)
            | (self.worker_id << self.WOKER_ID_SHIFT)
            | self.sequence
        )
        return new_id

    def __til_next_millis(self, last_timestamp) -> int:
        """等到下一毫秒."""
        timestamp = self.__gen_timestamp()
        while timestamp <= last_timestamp:
            timestamp = self.__gen_timestamp()
        return timestamp


ID_WORKER = IdWorker(1, 1, 0)


def get_id() -> str:
    """从 melobot 内部 id 生成器获得一个 id 值，不保证线程安全。

    :return: id 值
    """
    return str(ID_WORKER.get_id())


def get_rich_str(obj: object) -> str:
    """返回使用 rich 格式化的 object."""
    sio = io.StringIO()
    rich.print(obj, file=sio)
    return sio.getvalue().strip("\n")


def this_dir(*relative_path: str) -> str:
    """包内 py 脚本通过该方法可获得所在目录的绝对路径。提供参数，还可拼接路径。

    使用方法：

    .. code:: python

       this_dir() # 获取 ./ 的绝对路径
       this_dir('abc', 'env.toml') # 获取 ./abc/env.toml 的绝对路径

    一些注意事项如下：

    使用 `this_dir()`，只能这样导入：（导入语句后可以使用 as 子句）

    .. code:: python

       from melobot import this_dir
       # 或
       from melobot.base.tools import this_dir

    若 `B.py` 从 `A.py` 导入包含 :func:`this_dir` 调用的结构，
    导入前 :func:`this_dir` 必须已运行，而不能延迟求值

    `A.py` 中：

    .. code:: python

       class Foo:
           DIR = this_dir()
           LAMBDA_DIR = lambda: this_dir()
           GET_DIR = lambda: this_dir()
       OUTER_DIR = Foo.LAMBDA_DIR()  # Ok

    `B.py` 中：

    .. code:: python

       from .A import Foo, OUTER_DIR
       OUTER_DIR      # Ok
       Foo.DIR        # Ok
       Foo.GET_DIR()  # Error

    :param relative_path: 可用于拼接的路径部分
    :return: 拼接后的绝对路径
    """
    cur_finfo: inspect.FrameInfo | None = None
    cur_idx: int
    caller_path: str | None = None
    stacks = inspect.stack()

    for idx, finfo in enumerate(stacks):
        if finfo.function == "this_dir" and os.path.samefile(finfo.filename, __file__):
            cur_finfo, cur_idx = finfo, idx
    if cur_finfo is None:
        raise BotRuntimeError("this_dir 定位失败，请检查本函数使用方式是否正确")

    for idx, finfo in enumerate(stacks[cur_idx + 1 :]):
        if finfo.function == "<module>":
            for val in finfo.frame.f_locals.values():
                if val is __dir_inspector__:
                    caller_path = finfo.filename
                    break
            if caller_path is not None:
                break
    if caller_path is None:
        raise BotRuntimeError("this_dir 定位失败，请检查本函数使用方式是否正确")

    return str(
        pathlib.Path(caller_path).parent.joinpath(*relative_path).resolve(strict=True)
    )


__dir_inspector__ = this_dir


def to_async(func: Callable[[], T]) -> Callable[[], Coroutine[Any, Any, T]]:
    """异步包装函数

    将一个同步函数包装为异步函数，保留返回值。如果需要传参使用 partial 包裹。

    :param func: 需要转换的函数
    :return: 异步函数
    """

    async def wrapper():
        return func()

    return wrapper


def to_coro(func: Callable[[], T]) -> Coroutine[Any, Any, T]:
    """协程包装函数

    将一个同步函数包装为协程，保留返回值。如果需要传参使用 partial 包裹。

    :param func: 需要包装的函数
    :return: 协程
    """

    f = to_async(func)
    return f()


def to_task(obj: Callable[[], T] | Coroutine[Any, Any, T]) -> asyncio.Task[T]:
    """任务包装器

    将一个同步函数或协程包装为任务，保留返回值。如果需要传参使用 partial 包裹。

    :param obj: 同步函数或协程
    :return: 任务
    """
    if iscoroutine(obj):
        return asyncio.create_task(obj)
    elif iscoroutinefunction(obj):
        return asyncio.create_task(obj())
    else:
        return asyncio.create_task(to_coro(obj))  # type: ignore


def lock(callback: Callable[[], Coroutine[Any, Any, T1]]):
    """锁装饰器

    本方法作为异步函数的装饰器使用，可以为被装饰函数加锁。

    在获取锁冲突时，调用 `callback` 获得一个回调并执行。回调执行完毕后直接返回

    :param callback: 获取锁冲突时的回调
    :return:
       获取锁未冲突 -> 被装饰函数返回值

       获取锁冲突 -> 回调的返回值
    """
    alock = asyncio.Lock()
    if not callable(callback):
        raise BotBaseUtilsError(
            f"lock 装饰器的 callback 参数不可调用，callback 值为：{callback}"
        )

    def deco_func(
        func: Callable[P, Coroutine[Any, Any, T2]]
    ) -> Callable[P, Coroutine[Any, Any, T1 | T2]]:
        @wraps(func)
        async def wrapped_func(*args: Any, **kwargs: Any) -> T1 | T2:
            if alock.locked():
                cb = callback()
                if not iscoroutine(cb):
                    raise BotBaseUtilsError(
                        f"lock 装饰器的 callback 返回的不是协程，返回的回调为：{cb}"
                    )
                return await cb
            async with alock:
                return await func(*args, **kwargs)

        return wrapped_func

    return deco_func


def cooldown(
    busy_callback: Callable[[], Coroutine[Any, Any, T1]],
    cd_callback: Optional[Callable[[float], Coroutine[Any, Any, T2]]] = None,
    interval: float = 5,
):
    """冷却装饰器

    本方法作为异步函数的装饰器使用，可以为被装饰函数添加 cd 时间。

    如果被装饰函数已有一个在运行，此时调用 `busy_callback` 生成回调并执行。回调执行完毕后直接返回。

    如果被装饰函数没有在运行的，但在冷却完成前被调用：

       - `cd_callback` 不为空：使用 `cd_callback` 生成回调并执行。
       - `cd_callback` 为空， 被装饰函数持续等待，直至冷却结束再执行。

    :param busy_callback: 已运行时的回调
    :param cd_callback: 冷却时间未到的回调
    :param interval: 冷却时间
    :return:
       被装饰函数可执行 -> 被装饰函数返回值

       发生已运行冲突 -> `busy_callback` 的返回值

       冷却时间未到 -> `cd_callback` 的返回值
    """
    alock = asyncio.Lock()
    pre_finish_t = time.time() - interval - 1
    if not callable(busy_callback):
        raise BotBaseUtilsError(
            f"cooldown 装饰器的 busy_callback 参数不可调用，busy_callback 值为：{busy_callback}"
        )
    if cd_callback is not None and not callable(cd_callback):
        raise BotBaseUtilsError(
            f"cooldown 装饰器的 cd_callback 参数不可调用，cd_callback 值为：{cd_callback}"
        )

    def deco_func(
        func: Callable[P, Coroutine[Any, Any, T3]]
    ) -> Callable[P, Coroutine[Any, Any, T1 | T2 | T3]]:
        @wraps(func)
        async def wrapped_func(*args: Any, **kwargs: Any) -> T1 | T2 | T3:
            nonlocal pre_finish_t
            if alock.locked():
                busy_cb = busy_callback()
                if not iscoroutine(busy_cb):
                    raise BotBaseUtilsError(
                        f"cooldown 装饰器的 busy_callback 返回的不是协程，返回的回调为：{busy_cb}"
                    )
                return await busy_cb

            async with alock:
                duration = time.time() - pre_finish_t
                if duration > interval:
                    ret = await func(*args, **kwargs)
                    pre_finish_t = time.time()
                    return ret

                remain_t = interval - duration
                if cd_callback is not None:
                    cd_cb = cd_callback(remain_t)
                    if not iscoroutine(cd_cb):
                        raise BotBaseUtilsError(
                            f"cooldown 装饰器的 cd_callback 返回的不是协程，返回的回调为：{cd_cb}"
                        )
                    return await cd_cb
                else:
                    await asyncio.sleep(remain_t)
                    ret = await func(*args, **kwargs)
                    pre_finish_t = time.time()
                    return ret

        return wrapped_func

    return deco_func


def semaphore(callback: Callable[[], Coroutine[Any, Any, T1]], value: int = -1):
    """信号量装饰器

    本方法作为异步函数的装饰器使用，可以为被装饰函数添加信号量控制。

    在信号量无法立刻获取时，将调用 `callback` 获得回调并执行。回调执行完毕后直接返回。

    :param callback: 信号量无法立即获取的回调
    :param value: 信号量阈值
    :return:
       可以立即获取信号量 -> 被装饰函数返回值

       不能立即获取信号量 -> 回调的返回值
    """
    a_semaphore = asyncio.Semaphore(value)
    if not callable(callback):
        raise BotBaseUtilsError(
            f"semaphore 装饰器的 callback 参数不可调用，callback 值为：{callback}"
        )

    def deco_func(
        func: Callable[P, Coroutine[Any, Any, T2]]
    ) -> Callable[P, Coroutine[Any, Any, T1 | T2]]:
        @wraps(func)
        async def wrapped_func(*args: Any, **kwargs: Any) -> T1 | T2:
            if a_semaphore.locked():
                cb = callback()
                if not iscoroutine(cb):
                    raise BotBaseUtilsError(
                        f"semaphore 装饰器的 callback 返回的不是协程，返回的回调为：{cb}"
                    )
                return await cb
            async with a_semaphore:
                return await func(*args, **kwargs)

        return wrapped_func

    return deco_func


def timelimit(callback: Callable[[], Coroutine[Any, Any, T1]], timeout: float = 5):
    """时间限制装饰器

    本方法作为异步函数的装饰器使用，可以为被装饰函数添加超时控制。

    超时之后，调用 `callback` 获得回调并执行，同时取消原任务

    :param callback: 超时时的回调
    :param timeout: 超时时间, defaults to 5
    :return:
       被装饰函数未超时 -> 被装饰函数返回值

       被装饰函数超时 -> 回调的返回值
    """
    if not callable(callback):
        raise BotBaseUtilsError(
            f"timelimit 装饰器的 callback 参数不可调用，callback 值为：{callback}"
        )

    def deco_func(
        func: Callable[P, Coroutine[Any, Any, T2]]
    ) -> Callable[P, Coroutine[Any, Any, T1 | T2]]:
        @wraps(func)
        async def wrapped_func(*args: Any, **kwargs: Any) -> T1 | T2:
            try:
                return await asyncio.wait_for(func(*args, **kwargs), timeout)
            except asyncio.TimeoutError:
                cb = callback()
                if not iscoroutine(cb):
                    raise BotBaseUtilsError(
                        f"timelimit 装饰器的 callback 返回的不是协程，返回的回调为：{cb}"
                    )
                return await cb

        return wrapped_func

    return deco_func


def call_later(callback: Callable[[], None], delay: float):
    """同步函数延迟调度

    在指定的 `delay` 后调度一个 `callback` 执行。`callback` 应该是同步方法。

    :param callback: 同步函数
    :param delay: 多长时间后调度
    :return: :class:`asyncio.TimerHandle` 对象
    """
    return asyncio.get_running_loop().call_later(delay, callback)


def call_at(callback: Callable[[], None], timestamp: float):
    """同步函数指定时间调度

    在指定的时间戳调度一个 `callback` 执行。`callback` 应该是同步方法。`timestamp` <= 当前时刻回调立即执行

    :param callback: 同步函数
    :param timestamp: 在什么时刻调度
    :return: :class:`asyncio.TimerHandle` 对象
    """
    if timestamp <= time.time():
        return asyncio.get_running_loop().call_soon(callback)
    else:
        return asyncio.get_running_loop().call_later(timestamp - time.time(), callback)


def async_later(callback: Coroutine[Any, Any, T], delay: float) -> asyncio.Future[T]:
    """异步函数延迟调度（可自主选择是否等待）

    在指定的 `delay` 后调度一个 `callback` 执行。`callback` 是协程。

    返回一个 :class:`asyncio.Future` 对象，你可以选择等待或不等待。等待 :class:`asyncio.Future` 即是等待 `callback` 的返回值。

    注意：如果 `callback` 未完成就被取消，需要捕获 :class:`asyncio.CancelledError`。

    :param callback: 异步函数（可有返回值）
    :param delay: 多长时间后调度
    :return: :class:`asyncio.Future` 对象
    """

    async def async_cb(fut: asyncio.Future) -> Any:
        try:
            await asyncio.sleep(delay)
            res = await callback
            fut.set_result(res)
        except asyncio.CancelledError:
            callback.close()

    fut: asyncio.Future[Any] = asyncio.Future()
    asyncio.create_task(async_cb(fut))
    return fut


def async_at(callback: Coroutine[Any, Any, T], timestamp: float) -> asyncio.Future[T]:
    """异步函数指定时间调度（可自主选择是否等待）

    在指定的时间戳调度一个 `callback` 执行。`callback` 是协程。

    返回一个 :class:`asyncio.Future` 对象，你可以选择等待或不等待。等待 :class:`asyncio.Future` 即是等待 `callback` 的返回值。

    注意：如果 `callback` 未完成就被取消，需要捕获 :class:`asyncio.CancelledError`。

    :param callback: 异步函数（可有返回值）
    :param timestamp: 在什么时刻调度
    :return: :class:`asyncio.Future` 对象
    """
    if timestamp <= time.time():
        return async_later(callback, 0)
    else:
        return async_later(callback, timestamp - time.time())


def async_interval(
    callback: Callable[[], Coroutine[Any, Any, None]], interval: float
) -> asyncio.Task[None]:
    """异步函数间隔调度（类似 JavaScript 的 setInterval）

    每过时间间隔执行 `callback` 一次。`callback` 是返回协程的可调用对象（异步函数或 lambda 函数等）。

    返回一个 :class:`asyncio.Task` 对象，可使用该 task 取消调度过程。

    :param callback: 异步函数
    :param interval: 调度的间隔
    :return: :class:`asyncio.Task` 对象
    """

    async def interval_cb():
        try:
            while True:
                coro = callback()
                await asyncio.sleep(interval)
                await coro
        except asyncio.CancelledError:
            coro.close()

    t = asyncio.create_task(interval_cb())
    return t



import re

from ..base.abc import BotMatcher, BotMatcherError
from ..base.typing import Any, LogicMode


class AlwaysMatcher(BotMatcher):
    def __init__(self) -> None:
        super().__init__()

    def match(self, text: str) -> bool:
        return True


class StartMatcher(BotMatcher):
    """字符串起始匹配器"""

    def __init__(self, target: str | list[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串起始匹配器

        `target` 为字符串时，只进行一次起始匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行起始匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.target = target
        self.mode = mode

    def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text[: len(self.target)] == self.target
        else:
            res_seq = [text[: len(s)] == s for s in self.target]
            return LogicMode.seq_calc(self.mode, res_seq)


class ContainMatcher(BotMatcher):
    """字符串包含匹配器"""

    def __init__(self, target: str | list[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串包含匹配器

        `target` 为字符串时，只进行一次包含匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行包含匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.target = target
        self.mode = mode

    def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return self.target in text
        else:
            res_seq = [s in text for s in self.target]
            return LogicMode.seq_calc(self.mode, res_seq)


class EndMatcher(BotMatcher):
    """字符串结尾匹配器"""

    def __init__(self, target: str | list[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串结尾匹配器

        `target` 为字符串时，只进行一次结尾匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行结尾匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.target = target
        self.mode = mode

    def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text[-len(self.target) :] == self.target
        else:
            res_seq = [text[-len(s) :] == s for s in self.target]
            return LogicMode.seq_calc(self.mode, res_seq)


class FullMatcher(BotMatcher):
    """字符串全匹配器"""

    def __init__(self, target: str | list[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串全匹配器

        `target` 为字符串时，只进行一次全匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行全匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.target = target
        self.mode = mode

    def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text == self.target
        else:
            res_seq = [text == s for s in self.target]
            return LogicMode.seq_calc(self.mode, res_seq)


class RegexMatcher(BotMatcher):
    """字符串正则匹配器"""

    def __init__(self, regex_pattern: str, regex_flags: Any = 0) -> None:
        """初始化一个字符串正则匹配器

        :param regex_pattern: 正则 pattern
        :param regex_flags: 正则 flag，默认不使用
        """
        super().__init__()
        self.pattern = regex_pattern
        self.flag = regex_flags

    def match(self, text: str) -> bool:
        return len(re.findall(self.pattern, text, self.flag)) > 0



import asyncio
import http
import time

import websockets
import websockets.exceptions as wse

from ..base.abc import AbstractConnector, BotLife
from ..base.exceptions import get_better_exc
from ..base.tools import get_rich_str, to_task
from ..base.typing import TYPE_CHECKING, Any, ModuleType

if TYPE_CHECKING:
    import websockets.server

    from ..base.abc import BotAction


class ReverseWsConn(AbstractConnector):
    """反向 websocket 连接器

    .. admonition:: 注意
       :class: caution

       注意：在 melobot 中，反向 websocket 连接器会开启一个 ws 服务端。但是这个服务端只接受一个客户端连接，后续其他连接将被拒绝。
    """

    def __init__(
        self,
        listen_host: str,
        listen_port: int,
        cd_time: float = 0.2,
        allow_reconnect: bool = False,
    ) -> None:
        """初始化一个反向 websocket连接器

        :param listen_host: 监听的 host
        :param listen_port: 监听的 port
        :param cd_time: 行为操作冷却时间（用于防止风控）
        :param allow_reconnect: 是否允许客户端重新连接。默认为 `False`，即客户端连接后如果断连，将直接停止 bot；若为 `True`，则会等待客户端重连，等待时所有行为操作将阻塞
        """
        super().__init__(cd_time)
        #: 监听的 host
        self.host: str = listen_host
        #: 监听的 port
        self.port: int = listen_port
        #: 服务端对象
        self.server: "websockets.server.WebSocketServer"

        self._send_queue: asyncio.Queue["BotAction"] = asyncio.Queue()
        self._pre_send_time = time.time()
        self._server_close: asyncio.Future[Any]
        self._conn: "websockets.server.WebSocketServerProtocol"
        self._conn_requested = False
        self._request_lock = asyncio.Lock()
        self._conn_ready = asyncio.Event()
        self._allow_reconn = allow_reconnect
        self._reconn_flag = False

    async def _req_check(self, *args: Any, **kwargs: Any):
        """拦截握手请求，只允许一个客户端连接"""
        if not self._conn_requested:
            async with self._request_lock:
                if not self._conn_requested:
                    self._conn_requested = True
                    return None
        return (
            http.HTTPStatus.FORBIDDEN,
            [],
            b"Melobot already accepted the unique connection.\n",
        )

    async def _start(self) -> None:
        """启动服务"""
        self._server_close = asyncio.Future()
        async with websockets.serve(
            self._listen, self.host, self.port, process_request=self._req_check
        ) as server:
            self.server = server
            self.logger.info("连接器启动了 ws 服务，等待 ws 连接中")
            await self._server_close
        self.logger.info("ws 服务已关闭")

    def _close(self) -> None:
        """关闭服务"""
        if self._server_close.done():
            return
        else:
            self._server_close.set_result(True)

    def _restore_wait(self) -> None:
        """在客户端主动断开连接后，重置到可以接受新连接的状态"""
        self.logger.warning("前端主动断开连接，等待其重连中")
        self._conn_ready.clear()
        self._conn_requested = False
        del self._conn
        self._reconn_flag = True

    async def __aenter__(self) -> "ReverseWsConn":
        to_task(self._start())
        return self

    async def __aexit__(
        self, exc_type: type[Exception], exc_val: Exception, exc_tb: ModuleType
    ) -> bool:
        self._close()
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    async def _send(self, action: "BotAction") -> None:
        """发送一个 action 给连接器，实际上是先提交到 send_queue"""
        await self._ready_signal.wait()
        await self._conn_ready.wait()

        if self.slack:
            self.logger.debug(f"action {id(action)} 因 slack 状态被丢弃")
            return
        await self._send_queue.put(action)
        self.logger.debug(f"action {id(action)} 已成功加入发送队列")

    async def _send_queue_watch(self) -> None:
        """真正的发送方法。从 send_queue 提取 action 并按照一些处理步骤操作."""
        await self._ready_signal.wait()

        try:
            while True:
                action = await self._send_queue.get()
                await self._conn_ready.wait()
                self.logger.debug(
                    f"action {id(action)} 准备发送，结构如下：\n"
                    + get_rich_str(action.__dict__)
                )
                await self._bot_bus.emit(BotLife.ACTION_PRESEND, action, wait=True)
                self.logger.debug(f"action {id(action)} presend hook 已完成")
                action_str = action.flatten()
                wait_time = self.cd_time - (time.time() - self._pre_send_time)
                self.logger.debug(f"action {id(action)} 冷却等待：{wait_time}")
                await asyncio.sleep(wait_time)
                await self._conn.send(action_str)
                self.logger.debug(f"action {id(action)} 已发送")
                self._pre_send_time = time.time()
        except asyncio.CancelledError:
            self.logger.debug("连接器发送队列监视任务已被结束")
        except wse.ConnectionClosed:
            self.logger.error("连接器与前端的通信已经停止，无法再执行行为操作")

    async def _listen(self, ws: "websockets.server.WebSocketServerProtocol") -> None:
        """从前端接收一个事件，并处理"""
        self._conn = ws
        self.logger.info("连接器与前端建立了 ws 连接")
        await self._ready_signal.wait()
        self._conn_ready.set()
        if not self._reconn_flag:
            await self._bot_bus.emit(BotLife.CONNECTED)
            self.logger.debug("CONNECTED hook 已完成")

        try:
            while True:
                try:
                    raw_event = await self._conn.recv()
                    self.logger.debug(f"收到事件，未格式化的字符串：\n{raw_event}")
                    if raw_event == "":
                        continue
                    event = self._event_builder.build(raw_event)
                    self.logger.debug(
                        f"event {id(event)} 构建完成，结构：\n"
                        + get_rich_str(event.raw)
                    )
                    if event.is_resp_event():
                        to_task(self._resp_dispatcher.respond(event))  # type: ignore
                    else:
                        to_task(self._common_dispatcher.dispatch(event))  # type: ignore
                except wse.ConnectionClosed:
                    raise
                except Exception as e:
                    self.logger.error("bot 连接器监听任务抛出异常")
                    self.logger.error(f"异常点 raw_event：{raw_event}")
                    self.logger.error("异常回溯栈：\n" + get_better_exc(e))
                    self.logger.error("异常点局部变量：\n" + get_rich_str(locals()))
        except asyncio.CancelledError:
            self.logger.debug("连接器监听任务已停止")
        except wse.ConnectionClosed:
            self.logger.debug("连接器与前端的通信已经停止")
        finally:
            if self._server_close.done():
                return
            if not self._allow_reconn:
                self._close()
                return
            self._restore_wait()

    async def _alive_tasks(self) -> list[asyncio.Task]:
        async def alive():
            try:
                await self._server_close
            except asyncio.CancelledError:
                self._close()

        to_task(self._send_queue_watch())
        return [to_task(alive())]



"""MeloBot 是插件化管理、基于异步会话机制的 qq 机器人开发框架。

项目源码和文档，请参考：https://github.com/aicorein/melobot
"""

from .base import (
    LogicMode,
    PriorLevel,
    SessionRule,
    User,
    cooldown,
    lock,
    this_dir,
    timelimit,
)
from .bot import MeloBot, thisbot
from .context import (
    AttrSessionRule,
    finish,
    get_store,
    msg_args,
    msg_event,
    msg_text,
    pause,
    reply_finish,
    send,
    send_reply,
    send_wait,
)
from .io import ForwardWsConn, HttpConn, ReverseWsConn
from .meta import MetaInfo
from .plugin import BotPlugin
from .utils import (
    ArgFormatter,
    AtChecker,
    CmdParser,
    CmdParserGen,
    GroupMsgLvlChecker,
    MsgCheckerGen,
    PrivateMsgLvlChecker,
)

__version__ = MetaInfo().VER



import sys

sys.path.append("../src")

from melobot import BotPlugin, ForwardWsConn, MeloBot, msg_text, send_reply

bot1 = MeloBot("test1")
bot2 = MeloBot("test2")


class MyPlugin(BotPlugin):
    def __init__(self) -> None:
        super().__init__()

    @BotPlugin.on_start_match(".hello")
    async def hello(self) -> None:
        s = msg_text()
        await send_reply(s[6:])

    @bot1.on_loaded()
    async def tip(self) -> None:
        self.LOGGER.info("本插件已被加载！")


class MyPlugin2(BotPlugin):
    def __init__(self) -> None:
        super().__init__()

    @BotPlugin.on_start_match(".hello")
    async def hello(self) -> None:
        s = msg_text()
        await send_reply(s[6:] + "okokok")


if __name__ == "__main__":
    conn1 = ForwardWsConn("127.0.0.1", 8080)
    bot1.init(conn1, log_level="DEBUG")
    bot1.load_plugin(MyPlugin)

    conn2 = ForwardWsConn("127.0.0.1", 8081)
    bot2.init(conn2, log_level="DEBUG")
    bot2.load_plugin(MyPlugin2)

    MeloBot.start(bot1, bot2)



import asyncio
from asyncio import Future

from ..base.exceptions import get_better_exc
from ..base.tools import get_rich_str
from ..base.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..base.abc import AbstractConnector, BotAction
    from ..models.event import ResponseEvent
    from ..utils.logger import Logger


class BotResponder:
    """Bot 响应模块，是 action 发送方和 bot 连接模块的媒介。 提供 action 发送、响应回送功能."""

    def __init__(self) -> None:
        super().__init__()
        self._resp_table: dict[str, Future["ResponseEvent"]] = {}
        self.logger: "Logger"
        self._action_sender: "AbstractConnector"

        self._ready_signal = asyncio.Event()

    def _bind(self, logger: "Logger", connector: "AbstractConnector") -> None:
        self.logger = logger
        self._action_sender = connector

    def _set_ready(self) -> None:
        self._ready_signal.set()

    async def respond(self, resp: "ResponseEvent") -> None:
        await self._ready_signal.wait()

        try:
            self.logger.debug(
                f"收到 resp {id(resp)}，结构：\n" + get_rich_str(resp.raw)
            )
            if resp.id is None:
                return
            else:
                resp_fut = self._resp_table.get(resp.id)
                if resp_fut:
                    resp_fut.set_result(resp)
                    self._resp_table.pop(resp.id)
                else:
                    self.logger.error(f"收到了不匹配的携带 id 的响应：{resp.raw}")
        except asyncio.InvalidStateError:
            self.logger.warning(
                "等待响应事件的异步任务已被取消，这可能意味着连接器响应过慢，或任务设置的超时时间太短"
            )
            self._resp_table.pop(resp.id)  # type: ignore
        except Exception as e:
            self.logger.error("bot responder.dispatch 抛出异常")
            self.logger.error("异常点 resp_event：\n" + get_rich_str(resp))
            self.logger.error("异常回溯栈：\n" + get_better_exc(e))
            self.logger.error("异常点局部变量：\n" + get_rich_str(locals()))

    async def take_action(self, action: "BotAction") -> None:
        """响应器发送 action, 不等待响应."""
        await self._ready_signal.wait()
        await self._action_sender._send(action)
        return None

    async def take_action_wait(self, action: "BotAction") -> Future["ResponseEvent"]:
        """响应器发送 action，并返回一个 Future 用于等待响应."""
        await self._ready_signal.wait()
        fut: Future["ResponseEvent"] = Future()
        self._resp_table[action.resp_id] = fut  # type: ignore
        await self._action_sender._send(action)
        return fut



import asyncio
import time

import aiohttp
import aiohttp.web
from aiohttp.client_exceptions import ClientConnectorError

from ..base.abc import AbstractConnector, BotLife
from ..base.exceptions import get_better_exc
from ..base.tools import get_rich_str, to_task
from ..base.typing import TYPE_CHECKING, ModuleType

if TYPE_CHECKING:
    from ..base.abc import BotAction
    from ..models.event import ResponseEvent


class HttpConn(AbstractConnector):
    """HTTP 全双工连接器

    HTTP 全双工连接器将会同时开启 HTTP 服务端和客户端。

    客户端将会向 OneBot 实现程序发起 HTTP POST 请求用于传递行为操作，
    OneBot 实现程序将会向服务端发起 HTTP POST 请求用于事件上报。

    .. admonition:: 注意
       :class: caution

       HTTP 连接是无状态的，因此本连接器无法及时察觉 OneBot 实现程序掉线。只有在后续执行行为操作失败时，才会发现 OneBot 实现程序掉线。
    """

    def __init__(
        self,
        onebot_host: str,
        onebot_port: int,
        listen_host: str,
        listen_port: int,
        cd_time: float = 0.2,
        allow_reconnect: bool = False,
    ) -> None:
        """初始化一个 HTTP 全双工连接器

        :param onebot_host: onebot 实现程序 HTTP 服务的 host
        :param onebot_port: onebot 实现程序 HTTP 服务的 port
        :param listen_host: 此连接器服务端监听的 host
        :param listen_port: 此连接器服务端监听的 port
        :param allow_reconnect: 是否等待 OneBot 实现程序重新上线。默认为 `False`，即检测到 OneBot 实现程序掉线，将直接停止 bot；若为 `True`，则会等待 OneBot 实现程序重新上线，等待时所有行为操作将阻塞
        :param cd_time: 行为操作冷却时间（用于防止风控）
        """
        super().__init__(cd_time)
        #: onebot 实现程序提供服务的 base_url（形如：http://xxx:xxx）
        self.onebot_url = f"http://{onebot_host}:{onebot_port}"
        #: 本连接器服务端的 host
        self.host: str = listen_host
        #: 本连接器服务端的 port
        self.port: int = listen_port
        #: 本连接器服务端的站点对象
        self.serve_site: aiohttp.web.TCPSite
        #: 本连接器客户端的 HTTP 会话
        self.client_session: aiohttp.ClientSession

        self._send_queue: asyncio.Queue["BotAction"] = asyncio.Queue()
        self._pre_send_time = time.time()
        self._closed = asyncio.Event()
        self._close_lock = asyncio.Lock()
        self._onebot_onlined = asyncio.Event()
        self._allow_reconn = allow_reconnect

    async def _start(self) -> None:
        """启动连接器"""
        # 打开一个会话用于与 onebot 实现程序建立连接
        self.client_session = aiohttp.ClientSession()
        # 开启 HTTP 服务端
        app = aiohttp.web.Application()
        app.add_routes([aiohttp.web.post("/", self._listen)])
        runner = aiohttp.web.AppRunner(app)
        await runner.setup()
        self.serve_site = aiohttp.web.TCPSite(runner, self.host, self.port)
        await self.serve_site.start()
        self.logger.info("HTTP 通信就绪，等待 OneBot 实现程序上线中")
        await self._onebot_onlined.wait()
        self.logger.info("HTTP 双向通信已建立")
        await self._bot_bus.emit(BotLife.CONNECTED)
        self.logger.debug("CONNECTED hook 已完成")

    async def _close(self) -> None:
        """关闭连接器"""
        if self._closed.is_set():
            return
        async with self._close_lock:
            if self._closed.is_set():
                return
            # 关闭 HTTP 服务端
            await self.serve_site.stop()
            # 关闭与 onebot 实现程序的连接会话
            await self.client_session.close()
            self._closed.set()
            self.logger.info("HTTP 双向通信已停止")

    async def __aenter__(self) -> "HttpConn":
        to_task(self._start())
        return self

    async def __aexit__(
        self, exc_type: type[Exception], exc_val: Exception, exc_tb: ModuleType
    ) -> bool:
        await self._close()
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    async def _listen(self, request: aiohttp.web.Request) -> None:
        """从前端接收一个事件，并处理"""
        await self._ready_signal.wait()
        if not self._onebot_onlined.is_set():
            self._onebot_onlined.set()

        try:
            data: dict = await request.json()
            self.logger.debug("收到事件，未格式化的字典对象：\n" + get_rich_str(data))
            event = self._event_builder.build(data)
            self.logger.debug(
                f"event {id(event)} 构建完成，结构：\n" + get_rich_str(event.raw)
            )
            to_task(self._common_dispatcher.dispatch(event))  # type: ignore
        except Exception as e:
            self.logger.error("bot 连接器监听任务抛出异常")
            self.logger.error(f"异常点 raw_event：{data}")
            self.logger.error("异常回溯栈：\n" + get_better_exc(e))
            self.logger.error("异常点局部变量：\n" + get_rich_str(locals()))

    async def _send(self, action: "BotAction") -> None:
        """发送一个 action 给连接器，实际上是先提交到 send_queue"""
        await self._ready_signal.wait()

        if self.slack:
            self.logger.debug(f"action {id(action)} 因 slack 状态被丢弃")
            return
        await self._send_queue.put(action)
        self.logger.debug(f"action {id(action)} 已成功加入发送队列")

    async def _send_queue_watch(self) -> None:
        """真正的发送方法。从 send_queue 提取 action 并按照一些处理步骤操作."""

        async def take_action(action: "BotAction") -> None:
            try:
                _ = await self.client_session.post(
                    self.onebot_url + f"/{action.type}", json=action.params
                )
                if action.resp_id is None:
                    return
                raw_resp: dict = await _.json()
                resp: "ResponseEvent" = self._event_builder.build(raw_resp)  # type: ignore
                resp.id = action.resp_id
                to_task(self._resp_dispatcher.respond(resp))  # type: ignore
            except (RuntimeError, ClientConnectorError):
                if not self._allow_reconn:
                    self.logger.error("OneBot 实现程序已掉线，无法再执行行为操作")
                    await self._close()
                else:
                    self.logger.warning(
                        "OneBot 实现程序已掉线，正在等待 OneBot 实现程序重新上线"
                    )
                    self._onebot_onlined.clear()

        await self._ready_signal.wait()
        try:
            while True:
                action = await self._send_queue.get()
                await self._onebot_onlined.wait()
                self.logger.debug(
                    f"action {id(action)} 准备发送，结构如下：\n"
                    + get_rich_str(action.__dict__)
                )
                await self._bot_bus.emit(BotLife.ACTION_PRESEND, action, wait=True)
                self.logger.debug(f"action {id(action)} presend hook 已完成")
                wait_time = self.cd_time - (time.time() - self._pre_send_time)
                self.logger.debug(f"action {id(action)} 冷却等待：{wait_time}")
                await asyncio.sleep(wait_time)
                to_task(take_action(action))
                self.logger.debug(f"action {id(action)} 已发送")
                self._pre_send_time = time.time()
        except asyncio.CancelledError:
            self.logger.debug("连接器发送队列监视任务已被结束")

    async def _alive_tasks(self) -> list[asyncio.Task]:
        async def alive() -> None:
            try:
                await self._closed.wait()
            except asyncio.CancelledError:
                await self._close()

        to_task(self._send_queue_watch())
        return [to_task(alive())]



from flask import Flask, render_template

path = r"E:\projects\Python\git-proj\melobot\docs\build\html"
app = Flask(__name__, static_folder=path, static_url_path="/", template_folder=path)


@app.route("/")
def index():
    return render_template("index.html")


if __name__ == "__main__":
    app.run("127.0.0.1", 5000, debug=True)



from .cq import *
from .event import MessageEvent, MetaEvent, NoticeEvent, RequestEvent, ResponseEvent

__all__ = (
    "MessageEvent",
    "MetaEvent",
    "NoticeEvent",
    "RequestEvent",
    "ResponseEvent",
    "text_msg",
    "face_msg",
    "record_msg",
    "at_msg",
    "share_msg",
    "music_msg",
    "custom_music_msg",
    "image_msg",
    "reply_msg",
    "poke_msg",
    "xml_msg",
    "json_msg",
    "custom_msg_node",
    "refer_msg_node",
    "cq_filter_text",
    "cq_escape",
    "cq_anti_escape",
    "to_cq_arr",
    "to_cq_str",
    "custom_type_msg"
)



import re
from copy import deepcopy
from itertools import chain, zip_longest

from ..base.exceptions import BotActionError
from ..base.typing import (
    TYPE_CHECKING,
    Any,
    CQMsgDict,
    Literal,
    MsgNodeDict,
    Optional,
    Type,
)

if TYPE_CHECKING:
    from ..base.abc import BotAction


__all__ = (
    "at_msg",
    "cq_anti_escape",
    "cq_escape",
    "cq_filter_text",
    "custom_msg_node",
    "custom_music_msg",
    "face_msg",
    "image_msg",
    "json_msg",
    "music_msg",
    "poke_msg",
    "record_msg",
    "refer_msg_node",
    "reply_msg",
    "share_msg",
    "text_msg",
    "to_cq_arr",
    "to_cq_str",
    "xml_msg",
    "custom_type_msg"
)


def text_msg(
    text: str,
) -> CQMsgDict:
    """生成普通文本消息

    参数详细说明参考 onebot 标准：`纯文本 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#纯文本>`_

    :param text: 消息内容
    :return: onebot 标准中的消息段对象
    """
    return {"type": "text", "data": {"text": text}}


def face_msg(
    id: int,
) -> CQMsgDict:
    """生成 qq 表情消息

    参数详细说明参考 onebot 标准：`QQ 表情 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#qq-表情>`_

    :param id: qq 表情的 ID
    :return: onebot 标准中的消息段对象
    """
    return {"type": "face", "data": {"id": f"{id}"}}


def record_msg(
    file: str,
    magic: Literal[0, 1] = 0,
    cache: Literal[0, 1] = 1,
    proxy: Literal[0, 1] = 1,
    timeout: Optional[int] = None,
) -> CQMsgDict:
    """生成语音消息

    参数详细说明参考 onebot 标准：`语音 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#语音>`_

    :param file: 语音文件名
    :param magic: 是否使用变声
    :param cache: 是否使用已缓存文件
    :param proxy: 是否通过代理下载文件
    :param timeout: 超时时间，默认不启用
    :return: onebot 标准中的消息段对象
    """
    base: CQMsgDict = {
        "type": "record",
        "data": {
            "file": file,
            "magic": str(magic),
            "cache": str(cache),
            "proxy": str(proxy),
        },
    }
    if timeout:
        base["data"]["timeout"] = str(timeout)
    return base


def at_msg(qq: int | Literal["all"]) -> CQMsgDict:
    """生成艾特消息

    参数详细说明参考 onebot 标准：`某人 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#某人>`_

    :param qq: 艾特的 qq 号，"all" 表示艾特全体成员
    :return: onebot 标准中的消息段对象
    """
    base: CQMsgDict = {
        "type": "at",
        "data": {
            "qq": str(qq),
        },
    }
    return base


def share_msg(
    url: str,
    title: str,
    content: Optional[str] = None,
    image: Optional[str] = None,
) -> CQMsgDict:
    """生成链接分享消息

    参数详细说明参考 onebot 标准：`链接分享 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#链接分享>`_

    :param url: 链接的 url
    :param title: 消息的标题
    :param content: 消息的内容描述（可选）
    :param image: 消息的封面图 url（可选）
    :return: onebot 标准中的消息段对象
    """
    base: CQMsgDict = {
        "type": "share",
        "data": {
            "url": url,
            "title": title,
        },
    }
    if content:
        base["data"]["content"] = content
    if image:
        base["data"]["image"] = image
    return base


def music_msg(
    type: Literal["qq", "163", "xm"],
    id: str,
) -> CQMsgDict:
    """生成音乐分享消息

    参数详细说明参考 onebot 标准：`音乐分享 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#音乐分享->`_

    :param type: 音乐平台的类型（q 音、网易云、虾米）
    :param id: 歌曲 id
    :return: onebot 标准中的消息段对象
    """
    return {"type": "music", "data": {"type": type, "id": id}}


def custom_music_msg(
    url: str,
    audio: str,
    title: str,
    content: Optional[str] = None,
    image: Optional[str] = None,
) -> CQMsgDict:
    """生成音乐自定义分享 url

    参数详细说明参考 onebot 标准：`音乐自定义分享 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#音乐自定义分享->`_

    :param url: 点击后跳转目标 url
    :param audio: 音乐 url
    :param title: 标题
    :param content: 内容描述（可选）
    :param image: 封面图 url（可选）
    :return: onebot 标准中的消息段对象
    """
    base: CQMsgDict = {
        "type": "music",
        "data": {
            "type": "custom",
            "url": url,
            "audio": audio,
            "title": title,
        },
    }
    if content:
        base["data"]["content"] = content
    if image:
        base["data"]["image"] = image
    return base


def image_msg(
    file: str,
    type: Optional[Literal["flash"]] = None,
    cache: Literal[0, 1] = 1,
    proxy: Literal[0, 1] = 1,
    timeout: Optional[int] = None,
) -> CQMsgDict:
    """生成图片消息

    参数详细说明参考 onebot 标准：`图片 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#图片>`_

    :param file: 图片文件名或图片 base64 内容
    :param type: 图片类型
    :param cache: 是否使用已缓存的文件
    :param proxy: 是否通过代理下载文件
    :param timeout: 超时时间，默认不超时
    :return: onebot 标准中的消息段对象
    """
    base: CQMsgDict = {
        "type": "image",
        "data": {
            "file": file,
            "cache": str(cache),
            "proxy": str(proxy),
        },
    }
    if type:
        base["data"]["type"] = type
    if timeout:
        base["data"]["timeout"] = str(timeout)
    return base


def reply_msg(
    id: int,
) -> CQMsgDict:
    """生成回复消息

    参数详细说明参考 onebot 标准：`回复 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#回复>`_

    :param id: 消息 id
    :return: onebot 标准中的消息段对象
    """
    return {
        "type": "reply",
        "data": {
            "id": str(id),
        },
    }


def poke_msg(
    qq: int,
) -> CQMsgDict:
    """生成戳一戳消息

    .. admonition:: 提示
       :class: tip

       鉴于目前大多数实现 cq 协议的项目都采用了此参数规范，因此在 melobot 中，也采用此规范。虽然这种规范实际上不符合 onebot 标准

    :param qq: 戳的 qq 号
    :return: onebot 标准中的消息段对象
    """
    return {
        "type": "poke",
        "data": {
            "qq": str(qq),
        },
    }


def xml_msg(data: str) -> CQMsgDict:
    """生成 xml 消息

    参数详细说明参考 onebot 标准：`xml 消息 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#xml-消息>`_

    :param data: xml 内容
    :return: onebot 标准中的消息段对象
    """
    return {"type": "xml", "data": {"data": data}}


def json_msg(data: str) -> CQMsgDict:
    """生成 json 消息

    参数详细说明参考 onebot 标准：`json 消息 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#json-消息>`_

    :param data: json 内容
    :return: onebot 标准中的消息段对象
    """
    return {"type": "json", "data": {"data": data}}


def custom_msg_node(
    content: str | CQMsgDict | list[CQMsgDict],
    sendName: str,
    sendId: int,
    seq: Optional[list[CQMsgDict]] = None,
    useStd: bool = False,
) -> MsgNodeDict:
    """生成一个自定义合并消息转发结点

    .. admonition:: 提示
       :class: tip

       鉴于此方法涉及的消息构造，存在两种规范：
       `onebot 规范 <https://github.com/botuniverse/onebot-11/blob/master/message/segment.md#合并转发自定义节点>`_、
       `go-cq 规范 <https://docs.go-cqhttp.org/cqcode/#合并转发消息节点>`_，
       因此在 melobot 中，支持你使用 useStd 参数自定义选择哪种风格来构造

    :param content: 消息结点内容
    :param sendName: 消息结点标记的发送人名字
    :param sendId: 消息结点标记的发送人 qq 号
    :param seq: 消息 seq 号（一般来说你不需要使用这个）
    :param useStd: 消息段对象构造时是否遵循 onebot 标准，默认为否（使用 go-cq 风格）
    :return: onebot 标准中的消息段对象
    """
    if isinstance(content, str):
        _ = text_msg(content)
        if not isinstance(_, list):
            msgs = [_]
    elif isinstance(content, dict):
        msgs = [content]
    elif isinstance(content, list):
        temp = []
        for _ in content:
            if isinstance(_, list):
                temp.extend(_)
            else:
                temp.append(_)
        msgs = temp
    if not useStd:
        ret: MsgNodeDict = {
            "type": "node",
            "data": {"name": sendName, "uin": str(sendId), "content": msgs},
        }
    else:
        ret: MsgNodeDict = {  # type: ignore
            "type": "node",
            "data": {"user_id": sendId, "nickname": sendName, "content": msgs},
        }
    if seq:
        ret["data"]["seq"] = seq  # type: ignore
    return ret


def refer_msg_node(id: int) -> MsgNodeDict:
    """生成一个引用合并消息转发结点

    :param id: 消息 id
    :return: onebot 标准中的消息段对象
    """
    return {"type": "node", "data": {"id": str(id)}}


def custom_type_msg(type: str, params: dict[str, str]) -> CQMsgDict:
    """生成一个自定义消息段对象

    :param type: 消息段对象的类型标识
    :param params: 消息段的参数
    :return: 符合 onebot 格式，但不在 onebot 标准中的消息段对象
    """
    ret: CQMsgDict = {"type": type, "data": {}}
    for k, v in params.items():
        ret["data"][k] = str(v)
    return ret


def cq_filter_text(s: str) -> str:
    """cq 文本过滤函数

    可从 cq 字符串中过滤出纯文本消息部分

    :param s: cq 字符串
    :return: 纯文本消息部分
    """
    regex = re.compile(r"\[CQ:.*?\]")
    return regex.sub("", s)


def cq_escape(text: str) -> str:
    """cq 字符串特殊符号转义

    如：将 "&" 转义为 "&amp;"

    :param text: 需要转义的 cq 字符串
    :return: 转义特殊符号后的 cq 字符串
    """
    return (
        text.replace("&", "&amp;")
        .replace("[", "&#91;")
        .replace("]", "&#93;")
        .replace(",", "&#44;")
    )


def cq_anti_escape(text: str) -> str:
    """cq 字符串特殊符号逆转义

    如：将 "&amp;" 逆转义为 "&"

    :param text: 需要逆转义的 cq 字符串
    :return: 逆转义特殊符号后的 cq 字符串
    """
    return (
        text.replace("&#44;", ",")
        .replace("&#93;", "]")
        .replace("&#91;", "[")
        .replace("&amp;", "&")
    )


def to_cq_arr(s: str) -> list[CQMsgDict]:
    """将 cq 字符串转换为消息段对象列表

    :param s: cq 字符串
    :return: 消息段对象列表
    """

    def replace_func(m) -> str:
        s, e = m.regs[0]
        cq_texts.append(m.string[s:e])
        return "\u0000"

    cq_regex = re.compile(r"\[CQ:.*?\]")
    cq_texts: list[str] = []
    no_cq_str = cq_regex.sub(replace_func, s)
    pure_texts = map(
        lambda x: f"[CQ:text,text={x}]" if x != "" else x,
        no_cq_str.split("\u0000"),
    )
    _: str = "".join(
        chain.from_iterable(zip_longest(pure_texts, cq_texts, fillvalue=""))
    )

    cq_entity: list[str] = _.split("]")[:-1]
    content: list = []
    for e in cq_entity:
        __ = e.split(",")
        cq_type = __[0][4:]
        data: dict[str, float | int | str] = {}
        for param_pair in __[1:]:
            name, val = param_pair.split("=")
            if cq_type != "text":
                val = cq_anti_escape(val)
            if val.isdigit() or (len(val) >= 2 and val[0] == "-" and val[1:].isdigit()):
                data[name] = int(val)
                continue
            try:
                data[name] = float(val)
            except Exception:
                data[name] = val
        content.append({"type": cq_type, "data": data})
    return content


def to_cq_str(content: list[CQMsgDict]) -> str:
    """将消息段对象列表转换为 cq 字符串

    :param content: 消息段对象列表
    :return: cq 字符串
    """
    if isinstance(content, str):
        return content
    msgs: list[str] = []
    for item in content:
        if item["type"] == "text":
            msgs.append(item["data"]["text"])  # type: ignore
            continue
        s = f"[CQ:{item['type']}"
        for k, v in item["data"].items():
            s += f",{k}={cq_escape(str(v))}"
        s += "]"
        msgs.append(s)
    return "".join(msgs)


def _to_cq_str_action(action: "BotAction") -> "BotAction":
    def _format_msg_action(action: "BotAction") -> None:
        action.params["message"] = to_cq_str(action.params["message"])

    def _format_forward_action(action: "BotAction") -> None:
        for item in action.params["messages"]:
            if "id" in item["data"].keys():
                continue
            item["data"]["content"] = to_cq_str(item["data"]["content"])

    _action = deepcopy(action)
    if _action.type == "send_msg":
        _format_msg_action(_action)
    elif _action.type in ("send_private_forward_msg", "send_group_forward_msg"):
        _format_forward_action(_action)
    else:
        raise BotActionError("传入的 action 因类型不匹配，不可被 cq 序列化")
    return _action


def _get_cq(content: list[CQMsgDict], cq_type: str) -> list[CQMsgDict]:
    return [item for item in content if item["type"] == cq_type]


def _get_cq_params(
    content: list[CQMsgDict], cq_type: str, param: str, type: Optional[Type[Any]] = None
) -> list[Any]:
    res: list[Any] = []
    for item in content:
        if item["type"] == cq_type:
            val = item["data"].get(param)
            res.append(val)
    if type is not None:
        res = list(map(lambda x: type(x), res))
    return res



import traceback

from ..base.exceptions import ArgFormatInitError, ArgLackError, ArgVerifyFailed
from ..base.typing import (
    Any,
    Callable,
    Coroutine,
    ModuleType,
    Optional,
    ParseArgs,
    Void,
    VoidType,
)
from ..context.action import send


class FormatInfo:
    """命令参数格式化信息对象

    用于在命令参数格式化异常时传递信息。

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(
        self,
        src: str | VoidType,
        src_desc: Optional[str],
        src_expect: Optional[str],
        idx: int,
        exc_type: Exception,
        exc_tb: ModuleType,
        cmd_name: str,
    ) -> None:
        #: 命令参数格式化前的原值
        self.src = src
        #: 命令参数值的功能描述
        self.src_desc = src_desc
        #: 命令参数值的值期待描述
        self.src_expect = src_expect
        #: 命令参数值的顺序（从 0 开始索引）
        self.idx = idx
        #: 命令参数格式化异常时的异常对象
        self.exc_type = exc_type
        #: 命令参数格式化异常时的调用栈信息
        self.exc_tb = exc_tb
        #: 命令参数所属命令的命令名
        self.cmd_name = cmd_name


class ArgFormatter:
    """命令参数格式化器

    用于格式化命令解析器解析出的命令参数。
    """

    def __init__(
        self,
        convert: Optional[Callable[[str], Any]] = None,
        verify: Optional[Callable[[Any], bool]] = None,
        src_desc: Optional[str] = None,
        src_expect: Optional[str] = None,
        default: Any = Void,
        default_replace_flag: Optional[str] = None,
        convert_fail: Optional[
            Callable[[FormatInfo], Coroutine[Any, Any, None]]
        ] = None,
        verify_fail: Optional[Callable[[FormatInfo], Coroutine[Any, Any, None]]] = None,
        arg_lack: Optional[Callable[[FormatInfo], Coroutine[Any, Any, None]]] = None,
    ) -> None:
        """初始化一个命令参数格式化器

        :param convert: 类型转换方法，为空则不进行类型转换
        :param verify: 值验证方法，为空则不对值进行验证
        :param src_desc: 命令参数值的功能描述
        :param src_expect: 命令参数值的值期待描述
        :param default: 命令参数值的默认值（默认值 :class:`.Void` 表示无值，而不是 :obj:`None` 表达的空值）
        :param default_replace_flag: 命令参数使用默认值的标志
        :param convert_fail: 类型转换失败的回调，为空则使用默认回调
        :param verify_fail: 值验证失败的回调，为空则使用默认回调
        :param arg_lack: 参数缺失的回调，为空则使用默认回调
        """
        self.convert = convert
        self.verify = verify
        self.src_desc = src_desc
        self.src_expect = src_expect
        self.default = default
        self.default_replace_flag = default_replace_flag
        if self.default is Void and self.default_replace_flag is not None:
            raise ArgFormatInitError(
                "初始化参数格式化器时，使用“默认值替换标记”必须同时设置默认值"
            )

        self.convert_fail = convert_fail
        self.verify_fail = verify_fail
        self.arg_lack = arg_lack

    def _get_val(self, args: ParseArgs, idx: int) -> Any:
        if self.default is Void:
            if args.vals is None or len(args.vals) < idx + 1:
                raise ArgLackError
            else:
                return args.vals[idx]
        if args.vals is None:
            args.vals = [self.default]
        elif len(args.vals) < idx + 1:
            args.vals.append(self.default)
        return args.vals[idx]

    async def format(
        self,
        cmd_name: str,
        args: ParseArgs,
        idx: int,
    ) -> bool:
        # 格式化参数为对应类型的变量
        try:
            src = self._get_val(args, idx)
            if (
                self.default_replace_flag is not None
                and src == self.default_replace_flag
            ):
                src = self.default
            res = self.convert(src) if self.convert is not None else src

            if self.verify is None:
                pass
            elif self.verify(res):
                pass
            else:
                raise ArgVerifyFailed
            args.vals[idx] = res  # type: ignore
            return True
        except ArgVerifyFailed as e:
            info = FormatInfo(
                src, self.src_desc, self.src_expect, idx, e, traceback, cmd_name
            )
            if self.verify_fail:
                await self.verify_fail(info)
            else:
                await self._verify_fail_default(info)
            return False
        except ArgLackError as e:
            info = FormatInfo(
                Void, self.src_desc, self.src_expect, idx, e, traceback, cmd_name
            )
            if self.arg_lack:
                await self.arg_lack(info)
            else:
                await self._arglack_default(info)
            return False
        except Exception as e:
            info = FormatInfo(
                src, self.src_desc, self.src_expect, idx, e, traceback, cmd_name
            )
            if self.convert_fail:
                await self.convert_fail(info)
            else:
                await self._convert_fail_default(info)
            return False

    async def _convert_fail_default(self, info: FormatInfo) -> None:
        e_class = info.exc_type.__class__.__name__
        src = info.src.__repr__() if isinstance(info.src, str) else info.src
        tip = f"第 {info.idx+1} 个参数"
        tip += (
            f"（{info.src_desc}）无法处理，给定的值为：{src}。"
            if info.src_desc
            else f"给定的值 {src} 无法处理。"
        )
        tip += f"参数要求：{info.src_expect}。" if info.src_expect else ""
        tip += f"\n详细错误描述：[{e_class}] {info.exc_type}"
        tip = f"命令 {info.cmd_name} 参数格式化失败：\n" + tip
        await send(tip)

    async def _verify_fail_default(self, info: FormatInfo) -> None:
        src = info.src.__repr__() if isinstance(info.src, str) else info.src
        tip = f"第 {info.idx+1} 个参数"
        tip += (
            f"（{info.src_desc}）不符合要求，给定的值为：{src}。"
            if info.src_desc
            else f"给定的值 {src} 不符合要求。"
        )
        tip += f"参数要求：{info.src_expect}。" if info.src_expect else ""
        tip = f"命令 {info.cmd_name} 参数格式化失败：\n" + tip
        await send(tip)

    async def _arglack_default(self, info: FormatInfo) -> None:
        tip = f"第 {info.idx+1} 个参数"
        tip += f"（{info.src_desc}）缺失。" if info.src_desc else "缺失。"
        tip += f"参数要求：{info.src_expect}。" if info.src_expect else ""
        tip = f"命令 {info.cmd_name} 参数格式化失败：\n" + tip
        await send(tip)



from .abc import BotAction, BotChecker, BotEvent, BotLife, BotMatcher, SessionRule
from .tools import (
    async_at,
    async_interval,
    async_later,
    call_at,
    call_later,
    cooldown,
    lock,
    semaphore,
    this_dir,
    timelimit,
    to_async,
    to_coro,
)
from .typing import LogicMode, PriorLevel, User



import sys

sys.path.append("../../src/")

from datetime import datetime

from melobot import BotPlugin, ForwardWsConn, MeloBot, msg_event, send, send_reply

plugin = BotPlugin(__name__, "1.0.0")
OWNER_QID = 1574260633


@plugin.on_start_match(".hello")
async def say_hi() -> None:
    # send_reply 第一参数与 send 完全相同
    e = msg_event()
    if e.sender.id != 1574260633:
        await send_reply("你好~ 你不是我的主人哦")
        return
    # 接下来是机器人主人的处理逻辑
    await send(">w<")
    await send("主人好")

if __name__ == "__main__":
    bot = MeloBot(__name__)
    bot.init(ForwardWsConn("127.0.0.1", 8080))
    bot.load_plugin(plugin)
    bot.run()



import warnings

from ..base.abc import ActionArgs, BotAction, BotEvent
from ..base.exceptions import BotActionError, BotSessionError, DirectRetSignal
from ..base.tools import get_id
from ..base.typing import TYPE_CHECKING, CQMsgDict, Literal, MsgNodeDict, Optional
from ..models.cq import _to_cq_str_action, reply_msg, text_msg
from .session import SESSION_LOCAL
from .session import BotSessionManager as CtxManager

if TYPE_CHECKING:
    from ..models.event import ResponseEvent

__all__ = (
    "send_custom_msg",
    "send",
    "send_custom_forward",
    "send_forward",
    "msg_recall",
    "get_msg",
    "get_forward_msg",
    "get_image",
    "send_like",
    "group_kick",
    "group_ban",
    "group_whole_ban",
    "set_group_admin",
    "set_group_card",
    "set_group_name",
    "group_leave",
    "set_group_title",
    "set_friend_add",
    "set_group_add",
    "get_login_info",
    "get_stranger_info",
    "get_friend_list",
    "get_group_info",
    "get_group_list",
    "get_group_member_info",
    "get_group_member_list",
    "get_group_honor",
    "check_send_image",
    "check_send_record",
    "get_onebot_version",
    "get_onebot_status",
    "take_custom_action",
    "make_action",
    "send_wait",
    "send_reply",
    "finish",
    "reply_finish",
)


class MsgActionArgs(ActionArgs):
    """消息 action 信息构造类."""

    def __init__(
        self,
        msgs: list[CQMsgDict],
        isPrivate: bool,
        userId: Optional[int] = None,
        groupId: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.type = "send_msg"
        if isPrivate:
            self.params = {
                "message_type": "private",
                "user_id": userId,
                "message": msgs,
            }
        else:
            self.params = {
                "message_type": "group",
                "user_id": userId,
                "group_id": groupId,
                "message": msgs,
            }


def _process_msg(content: str | CQMsgDict | list[CQMsgDict]) -> list[CQMsgDict]:
    """将多种可能的消息格式，统一转换为 cq 消息列表."""
    if isinstance(content, str):
        _ = text_msg(content)
        if not isinstance(_, list):
            msgs = [_]
    elif isinstance(content, dict):
        msgs = [content]
    elif isinstance(content, list):
        temp = []
        for _ in content:
            if isinstance(_, list):
                temp.extend(_)
            else:
                temp.append(_)
        msgs = temp
    else:
        raise BotActionError("content 参数类型不正确，无法封装")
    return msgs


@CtxManager._activate
async def send_custom_msg(
    content: str | CQMsgDict | list[CQMsgDict],
    isPrivate: bool,
    userId: Optional[int] = None,
    groupId: Optional[int] = None,
    cq_str: bool = False,
    wait: bool = False,
    auto: bool = True,
) -> BotAction:
    """发送消息（自定义发送目标）

    .. admonition:: 小技巧
       :class: note

       当启用了 `cq_str` 选项，且 `content` 参数为字符串时，
       `content` 字符串将不会被处理为消息段对象，此时字符串中的 cq 码将会直接生效，而不是被转义。

    .. admonition:: 警告
       :class: attention

       这个小技巧可以让你发送有效的 cq 码字符串。但是存在潜在的安全问题：
       如果将用户输入作为 cq 码字符串的一部分发送出去，这将会造成“注入攻击”！
       用户可以构造包含恶意图片、语音的 cq 码，让 bot 发送。

       任何时候启用 `cq_str` 选项，如需用到用户输入，务必校验。

    此接口合并了 onebot 标准中的私聊消息发送、群聊消息发送接口。

    :param content: 发送内容
    :param isPrivate: 是否是私聊
    :param userId: 如果是私聊，传入目标 qq 号；群聊置空即可
    :param groupId: 如果是群聊，传入群号；私聊置空即可
    :param cq_str: 是否以 cq 字符串发送（默认格式是消息段对象)
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    if isPrivate and userId is None:
        raise BotActionError("为私聊时，构建 action 必须提供 userId 参数")
    if not isPrivate and groupId is None:
        raise BotActionError("为群聊时，构建 action 必须提供 groupId 参数")
    action = BotAction(
        MsgActionArgs(_process_msg(content), isPrivate, userId, groupId),
        resp_id=get_id() if wait else None,
        ready=auto,
    )
    if cq_str:
        action = _to_cq_str_action(action)
    return action


@CtxManager._activate
async def send(
    content: str | CQMsgDict | list[CQMsgDict],
    cq_str: bool = False,
    wait: bool = False,
    auto: bool = True,
) -> BotAction:
    """发送消息（在当前会话下自动定位发送目标）

    .. admonition:: 小技巧
       :class: note

       当启用了 `cq_str` 选项，且 `content` 参数为字符串时，
       `content` 字符串将不会被处理为消息段对象，此时字符串中的 cq 码将会直接生效，而不是被转义。

    .. admonition:: 警告
       :class: attention

       这个小技巧可以让你发送有效的 cq 码字符串。但是存在潜在的安全问题：
       如果将用户输入作为 cq 码字符串的一部分发送出去，这将会造成“注入攻击”！
       用户可以构造包含恶意图片、语音的 cq 码，让 bot 发送。

       任何时候启用 `cq_str` 选项，如需用到用户输入，务必校验。

    :param content: 发送内容（可以是文本、消息段对象、消息段对象列表）
    :param cq_str: 是否以 cq 字符串发送（默认格式是消息段对象)
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    try:
        session = SESSION_LOCAL
        action = BotAction(
            MsgActionArgs(
                _process_msg(content),
                session.event.is_private(),
                session.event.sender.id,
                session.event.group_id,
            ),
            resp_id=get_id() if wait else None,
            ready=auto,
        )
        if cq_str:
            action = _to_cq_str_action(action)
        return action
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")


class ForwardMsgActionArgs(ActionArgs):
    """转发消息 action 信息构造类."""

    def __init__(
        self,
        msgs: list[MsgNodeDict],
        isPrivate: bool,
        userId: Optional[int] = None,
        groupId: Optional[int] = None,
    ) -> None:
        super().__init__()
        if isPrivate:
            self.type = "send_private_forward_msg"
            self.params = {"user_id": userId, "messages": msgs, "auto_escape": True}
        else:
            self.type = "send_group_forward_msg"
            self.params = {"group_id": groupId, "messages": msgs, "auto_escape": True}


@CtxManager._activate
async def send_custom_forward(
    msgNodes: list[MsgNodeDict],
    isPrivate: bool,
    userId: Optional[int] = None,
    groupId: Optional[int] = None,
    cq_str: bool = False,
    wait: bool = False,
    auto: bool = True,
) -> BotAction:
    """发送转发消息（自定义发送目标）

    .. admonition:: 小技巧
       :class: note

       当启用了 `cq_str` 选项，且 `content` 参数为字符串时，
       `content` 字符串将不会被处理为消息段对象，此时字符串中的 cq 码将会直接生效，而不是被转义。

    .. admonition:: 警告
       :class: attention

       这个小技巧可以让你发送有效的 cq 码字符串。但是存在潜在的安全问题：
       如果将用户输入作为 cq 码字符串的一部分发送出去，这将会造成“注入攻击”！
       用户可以构造包含恶意图片、语音的 cq 码，让 bot 发送。

       任何时候启用 `cq_str` 选项，如需用到用户输入，务必校验。

    :param msgNodes: 消息结点列表
    :param isPrivate: 是否是私聊
    :param userId: 如果是私聊，传入目标 qq 号；群聊置空即可
    :param groupId: 如果是群聊，传入群号；私聊置空即可
    :param cq_str: 是否以 cq 字符串发送（默认格式是消息段对象)
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    action = BotAction(
        ForwardMsgActionArgs(msgNodes, isPrivate, userId, groupId),
        resp_id=get_id() if wait else None,
        ready=auto,
    )
    if cq_str:
        action = _to_cq_str_action(action)
    return action


@CtxManager._activate
async def send_forward(
    msgNodes: list[MsgNodeDict],
    cq_str: bool = False,
    wait: bool = False,
    auto: bool = True,
) -> BotAction:
    """发送转发消息（在当前会话下自动定位发送目标）

    .. admonition:: 小技巧
       :class: note

       当启用了 `cq_str` 选项，且 `content` 参数为字符串时，
       `content` 字符串将不会被处理为消息段对象，此时字符串中的 cq 码将会直接生效，而不是被转义。

    .. admonition:: 警告
       :class: attention

       这个小技巧可以让你发送有效的 cq 码字符串。但是存在潜在的安全问题：
       如果将用户输入作为 cq 码字符串的一部分发送出去，这将会造成“注入攻击”！
       用户可以构造包含恶意图片、语音的 cq 码，让 bot 发送。

       任何时候启用 `cq_str` 选项，如需用到用户输入，务必校验。

    :param msgNodes: 消息结点列表
    :param cq_str: 是否以 cq 字符串发送（默认格式是消息段对象)
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    try:
        session = SESSION_LOCAL
        action = BotAction(
            ForwardMsgActionArgs(
                msgNodes,
                session.event.is_private(),
                session.event.sender.id,
                session.event.group_id,
            ),
            resp_id=get_id() if wait else None,
            ready=auto,
        )
        if cq_str:
            action = _to_cq_str_action(action)
        return action
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")


class MsgDelActionArgs(ActionArgs):
    """撤回消息 action 信息构造类."""

    def __init__(self, msgId: int) -> None:
        super().__init__()
        self.type = "delete_msg"
        self.params = {
            "message_id": msgId,
        }


@CtxManager._activate
async def msg_recall(msgId: int, wait: bool = False, auto: bool = True) -> BotAction:
    """撤回消息

    :param msgId: 消息 id
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        MsgDelActionArgs(msgId), resp_id=get_id() if wait else None, ready=auto
    )


class GetMsgActionArgs(ActionArgs):
    """消息信息获取 action 信息构造类."""

    def __init__(self, msgId: int) -> None:
        super().__init__()
        self.type = "get_msg"
        self.params = {"message_id": msgId}


@CtxManager._activate
async def get_msg(msgId: int, wait: bool = True, auto: bool = True) -> BotAction:
    """获取消息详细信息

    :param msgId: 消息 id
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetMsgActionArgs(msgId), resp_id=get_id() if wait else None, ready=auto
    )


class getForwardActionArgs(ActionArgs):
    """转发消息获取 action 信息构造类."""

    def __init__(self, forwardId: str) -> None:
        super().__init__()
        self.type = "get_forward_msg"
        self.params = {"message_id": forwardId}


@CtxManager._activate
async def get_forward_msg(
    forwardId: str, wait: bool = True, auto: bool = True
) -> BotAction:
    """获取转发消息的详细信息

    :param forwardId: 转发 id
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        getForwardActionArgs(forwardId), resp_id=get_id() if wait else None, ready=auto
    )


class getImageActionArgs(ActionArgs):
    """获取图片信息 action 信息构造类."""

    def __init__(self, fileName: str) -> None:
        super().__init__()
        self.type = "get_image"
        self.params = {"file": fileName}


@CtxManager._activate
async def get_image(fileName: str, wait: bool = True, auto: bool = True) -> BotAction:
    """获取图片信息

    :param fileName: 图片文件名
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        getImageActionArgs(fileName), resp_id=get_id() if wait else None, ready=auto
    )


class SendLikeActionArgs(ActionArgs):
    """发送好友赞 action 信息构造类"""

    def __init__(self, userId: int, times: int = 1) -> None:
        super().__init__()
        self.type = "send_like"
        self.params = {"user_id": userId, "times": times}


@CtxManager._activate
async def send_like(
    userId: int, times: int = 1, wait: bool = False, auto: bool = True
) -> BotAction:
    """发送好友赞

    :param userId: qq 号
    :param times: 赞的数量，默认为 1，每天最多 10
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        SendLikeActionArgs(userId, times),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GroupKickActionArgs(ActionArgs):
    """群组踢人 action 信息构造类."""

    def __init__(self, groupId: int, userId: int, laterReject: bool = False) -> None:
        super().__init__()
        self.type = "set_group_kick"
        self.params = {
            "group_id": groupId,
            "user_id": userId,
            "reject_add_request": laterReject,
        }


@CtxManager._activate
async def group_kick(
    groupId: int,
    userId: int,
    laterReject: bool = False,
    wait: bool = False,
    auto: bool = True,
) -> BotAction:
    """群组踢人

    :param groupId: 群号
    :param userId: 被踢的 qq 号
    :param laterReject: 是否拒绝此人再次加群
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GroupKickActionArgs(groupId, userId, laterReject),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GroupBanActionArgs(ActionArgs):
    """群组禁言 action 信息构造类."""

    def __init__(self, groupId: int, userId: int, duration: int) -> None:
        super().__init__()
        self.type = "set_group_ban"
        self.params = {
            "group_id": groupId,
            "user_id": userId,
            "duration": duration,
        }


@CtxManager._activate
async def group_ban(
    groupId: int, userId: int, duration: int, wait: bool = False, auto: bool = True
) -> BotAction:
    """群组禁言

    :param groupId: 群号
    :param userId: 禁言的 qq 号
    :param duration: 禁言时长，为 0 则表示取消禁言
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GroupBanActionArgs(groupId, userId, duration),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GroupWholeBanActionArgs(ActionArgs):
    """群组全员禁言 action 信息构造类."""

    def __init__(self, groupId: int, enable: bool) -> None:
        super().__init__()
        self.type = "set_group_whole_ban"
        self.params = {"group_id": groupId, "enable": enable}


@CtxManager._activate
async def group_whole_ban(
    groupId: int, enable: bool, wait: bool = False, auto: bool = True
) -> BotAction:
    """群组全员禁言

    :param groupId: 群号
    :param enable: 是则禁言，否则取消禁言
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GroupWholeBanActionArgs(groupId, enable),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class SetGroupAdminActionArgs(ActionArgs):
    """设置群管理员 action 信息构造类."""

    def __init__(self, groupId: int, userId: int, enable: bool) -> None:
        super().__init__()
        self.type = "set_group_admin"
        self.params = {"group_id": groupId, "user_id": userId, "enable": enable}


@CtxManager._activate
async def set_group_admin(
    groupId: int, userId: int, enable: bool, wait: bool = False, auto: bool = True
) -> BotAction:
    """设置群管理员

    :param groupId: 群号
    :param userId: 设置的 qq 号
    :param enable: 是则设置，否则取消
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        SetGroupAdminActionArgs(groupId, userId, enable),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class SetGroupCardActionArgs(ActionArgs):
    """设置群名片 action 信息构造类."""

    def __init__(self, groupId: int, userId: int, card: str) -> None:
        super().__init__()
        self.type = "set_group_card"
        self.params = {"group_id": groupId, "user_id": userId, "card": card}


@CtxManager._activate
async def set_group_card(
    groupId: int, userId: int, card: str, wait: bool = False, auto: bool = True
) -> BotAction:
    """设置群名片

    :param groupId: 群号
    :param userId: 设置的 qq 号
    :param card: 新名片内容
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        SetGroupCardActionArgs(groupId, userId, card),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class SetGroupNameActionArgs(ActionArgs):
    """设置群名 action 信息构造类."""

    def __init__(self, groupId: int, name: str) -> None:
        super().__init__()
        self.type = "set_group_name"
        self.params = {"group_id": groupId, "group_name": name}


@CtxManager._activate
async def set_group_name(
    groupId: int, name: str, wait: bool = False, auto: bool = True
) -> BotAction:
    """设置群名

    :param groupId: 群号
    :param name: 新群名
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        SetGroupNameActionArgs(groupId, name),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GroupLeaveActionArgs(ActionArgs):
    """退出群组 action 信息构造类."""

    def __init__(self, groupId: int, isDismiss: bool) -> None:
        super().__init__()
        self.type = "set_group_leave"
        self.params = {"group_id": groupId, "is_dismiss": isDismiss}


@CtxManager._activate
async def group_leave(
    groupId: int, isDismiss: bool, wait: bool = False, auto: bool = True
) -> BotAction:
    """退出群

    :param groupId: 群号
    :param isDismiss: 是否解散群（仅在 bot 为群主时可用）
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GroupLeaveActionArgs(groupId, isDismiss),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class SetGroupTitleActionArgs(ActionArgs):
    """设置群特殊头衔 action 信息构造类."""

    def __init__(
        self,
        groupId: int,
        userId: int,
        title: str,
        duration: int = -1,
    ) -> None:
        super().__init__()
        self.type = "set_group_special_title"
        self.params = {
            "group_id": groupId,
            "user_id": userId,
            "special_title": title,
            "duration": duration,
        }


@CtxManager._activate
async def set_group_title(
    groupId: int,
    userId: int,
    title: str,
    duration: int = -1,
    wait: bool = False,
    auto: bool = True,
) -> BotAction:
    """设置群特殊头衔

    :param groupId: 群号
    :param userId: 设置的 qq 号
    :param title: 头衔名
    :param duration: 生效时间，为 -1 则为无限期
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        SetGroupTitleActionArgs(groupId, userId, title, duration),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class SetFriendAddActionArgs(ActionArgs):
    def __init__(self, addFlag: str, approve: bool, remark: str) -> None:
        """处理加好友请求 action 信息构造类."""
        super().__init__()
        self.type = "set_friend_add_request"
        self.params = {"flag": addFlag, "approve": approve, "remark": remark}


@CtxManager._activate
async def set_friend_add(
    addFlag: str, approve: bool, remark: str, wait: bool = False, auto: bool = True
) -> BotAction:
    """处理加好友请求

    :param addFlag: 好友添加 flag，对应 :attr:`~.RequestEvent.req_flag` 属性
    :param approve: 是否通过
    :param remark: 添加后的好友备注（仅用于通过请求后）
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        SetFriendAddActionArgs(addFlag, approve, remark),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class SetGroupAddActionArgs(ActionArgs):
    """处理加群请求 action 信息构造类."""

    def __init__(
        self,
        addFlag: str,
        addType: Literal["add", "invite"],
        approve: bool,
        reason: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.type = "set_group_add_request"
        self.params = {
            "flag": addFlag,
            "sub_type": addType,
            "approve": approve,
        }
        if reason:
            self.params["reason"] = reason


@CtxManager._activate
async def set_group_add(
    addFlag: str,
    addType: Literal["add", "invite"],
    approve: bool,
    rejectReason: Optional[str] = None,
    wait: bool = False,
    auto: bool = True,
) -> BotAction:
    """处理加群请求（只有 bot 是群管理时有用）

    :param addFlag: 加群 flag，对应 :attr:`~.RequestEvent.req_flag` 属性
    :param addType: 加群类型，对应 :attr:`~.RequestEvent.group_req_type` 属性
    :param approve: 是否通过
    :param rejectReason: 如果不通过的原因回复
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        SetGroupAddActionArgs(addFlag, addType, approve, rejectReason),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GetLoginInfoActionArgs(ActionArgs):
    """获取登录号信息 action 信息构造类."""

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_login_info"
        self.params = {}


@CtxManager._activate
async def get_login_info(wait: bool = True, auto: bool = True) -> BotAction:
    """获得 bot 登录号信息

    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetLoginInfoActionArgs(), resp_id=get_id() if wait else None, ready=auto
    )


class GetStrangerInfoActionArgs(ActionArgs):
    """获取陌生人信息 action 信息构造类."""

    def __init__(self, userId: int, noCache: bool) -> None:
        super().__init__()
        self.type = "get_stranger_info"
        self.params = {"user_id": userId, "no_cache": noCache}


@CtxManager._activate
async def get_stranger_info(
    userId: int, noCache: bool, wait: bool = True, auto: bool = True
) -> BotAction:
    """获取陌生人信息，也可以对好友使用

    :param userId: qq 号
    :param noCache: 是否不使用缓存
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetStrangerInfoActionArgs(userId, noCache),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GetFriendlistActionArgs(ActionArgs):
    """获取好友列表 action 信息构造类."""

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_friend_list"
        self.params = {}


@CtxManager._activate
async def get_friend_list(wait: bool = True, auto: bool = True) -> BotAction:
    """获取好友列表

    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetFriendlistActionArgs(), resp_id=get_id() if wait else None, ready=auto
    )


class GetGroupInfoActionArgs(ActionArgs):
    """获取群信息 action 信息构造类."""

    def __init__(self, groupId: int, noCache: bool) -> None:
        super().__init__()
        self.type = "get_group_info"
        self.params = {"group_id": groupId, "no_cache": noCache}


@CtxManager._activate
async def get_group_info(
    groupId: int, noCache: bool, wait: bool = True, auto: bool = True
) -> BotAction:
    """获取群信息，可以是未加入的群聊

    :param groupId: 群号
    :param noCache: 是否不使用缓存
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetGroupInfoActionArgs(groupId, noCache),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GetGrouplistActionArgs(ActionArgs):
    """获取群列表 action 信息构造类."""

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_group_list"
        self.params = {}


@CtxManager._activate
async def get_group_list(wait: bool = True, auto: bool = True) -> BotAction:
    """获取群列表。

    可能返回的建群时间都是 0，这是不准确的。准确的时间可以通过 :meth:`get_group_info` 获得.

    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetGrouplistActionArgs(), resp_id=get_id() if wait else None, ready=auto
    )


class GetGroupMemberInfoActionArgs(ActionArgs):
    """获取群成员信息 action 信息构造类."""

    def __init__(self, groupId: int, userId: int, noCache: bool) -> None:
        super().__init__()
        self.type = "get_group_member_info"
        self.params = {"group_id": groupId, "user_id": userId, "no_cache": noCache}


@CtxManager._activate
async def get_group_member_info(
    groupId: int, userId: int, noCache: bool, wait: bool = True, auto: bool = True
) -> BotAction:
    """获取群成员信息

    :param groupId: 群号
    :param userId: qq 号
    :param noCache: 是否不使用缓存
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetGroupMemberInfoActionArgs(groupId, userId, noCache),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GetGroupMemberlistActionArgs(ActionArgs):
    """获取群成员列表 action 信息构造类."""

    def __init__(self, groupId: int, noCache: bool) -> None:
        super().__init__()
        self.type = "get_group_member_list"
        self.params = {"group_id": groupId, "no_cache": noCache}


@CtxManager._activate
async def get_group_member_list(
    groupId: int, noCache: bool, wait: bool = True, auto: bool = True
) -> BotAction:
    """获取群成员列表

    :param groupId: 群号
    :param noCache: 是否不使用缓存
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetGroupMemberlistActionArgs(groupId, noCache),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class GetGroupHonorActionArgs(ActionArgs):
    """获取群荣誉信息 action 信息构造类."""

    def __init__(
        self,
        groupId: int,
        type: Literal[
            "talkative", "performer", "legend", "strong_newbie", "emotion", "all"
        ],
    ) -> None:
        super().__init__()
        self.type = "get_group_honor_info"
        self.params = {"group_id": groupId, "type": type}


@CtxManager._activate
async def get_group_honor(
    groupId: int,
    type: Literal[
        "talkative", "performer", "legend", "strong_newbie", "emotion", "all"
    ],
    wait: bool = True,
    auto: bool = True,
) -> BotAction:
    """获取群荣誉信息

    详细说明参考：
    `获取群荣誉信息 <https://github.com/botuniverse/onebot-11/blob/master/api/public.md#get_group_honor_info-获取群荣誉信息>`_

    :param groupId: 群号
    :param type: 荣誉类型
    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetGroupHonorActionArgs(groupId, type),
        resp_id=get_id() if wait else None,
        ready=auto,
    )


class CheckSendImageActionArgs(ActionArgs):
    """检查是否可以发送图片 action 信息构造类."""

    def __init__(self) -> None:
        super().__init__()
        self.type = "can_send_image"
        self.params = {}


@CtxManager._activate
async def check_send_image(wait: bool = True, auto: bool = True) -> BotAction:
    """检查是否可以发送图片

    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        CheckSendImageActionArgs(), resp_id=get_id() if wait else None, ready=auto
    )


class CheckSendRecordActionArgs(ActionArgs):
    """检查是否可以发送语音 action 信息构造类."""

    def __init__(self) -> None:
        super().__init__()
        self.type = "can_send_image"
        self.params = {}


@CtxManager._activate
async def check_send_record(wait: bool = True, auto: bool = True) -> BotAction:
    """检查是否可以发送语音

    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        CheckSendRecordActionArgs(), resp_id=get_id() if wait else None, ready=auto
    )


class GetCqVersionActionArgs(ActionArgs):
    """获取 onebot 实现版本 action 信息构造类."""

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_version_info"
        self.params = {}


@CtxManager._activate
async def get_onebot_version(wait: bool = True, auto: bool = True) -> BotAction:
    """获取 onebot 实现项目的版本

    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetCqVersionActionArgs(), resp_id=get_id() if wait else None, ready=auto
    )


class GetCqStatusActionArgs(ActionArgs):
    """获取 onebot 实现 状态 action 信息构造类."""

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_status"
        self.params = {}


@CtxManager._activate
async def get_onebot_status(wait: bool = True, auto: bool = True) -> BotAction:
    """获取 onebot 实现项目的状态

    :param wait: 是否等待这个行为的响应
    :param auto: 是否自动发送
    :return:
       `auto=False` -> :class:`.BotAction` 对象

       `auto=True, wait=True` -> :class:`.ResponseEvent` 对象

       `auto=True, wait=False` -> :obj:`None`
    """
    return BotAction(
        GetCqStatusActionArgs(), resp_id=get_id() if wait else None, ready=auto
    )


@CtxManager._activate
async def take_custom_action(
    action: BotAction,
) -> BotAction:
    """发送给定的自定义行为

    :param action: 行为对象
    :return:
       若此 action 指定等待 -> :class:`.ResponseEvent` 对象

       若此 action 未指定等待 -> :obj:`None`
    """
    action._ready = True
    return action


async def make_action(
    type: str, params: dict, need_resp: bool, trigger: Optional[BotEvent] = None
) -> BotAction:
    """创建一个自定义行为

    :param type: 行为的类型
    :param params: 行为的附加参数
    :param need_resp: 是否需要等待这个行为
    :param trigger: 行为的触发事件（一般不用填。如果不需要跟踪触发事件，则不填）
    :return: 行为对象
    """
    args = ActionArgs()
    args.type, args.params = type, params
    return BotAction(args, get_id() if need_resp else None, trigger, ready=False)


async def send_wait(
    content: str | CQMsgDict | list[CQMsgDict],
    cq_str: bool = False,
    overtime: Optional[int] = None,
) -> None:
    """发送一条消息然后暂停当前会话（在当前会话下自动定位发送目标）

    会话应该是可暂停的，这意味着该会话应该拥有会话规则。

    .. admonition:: 小技巧
       :class: note

       当启用了 `cq_str` 选项，且 `content` 参数为字符串时，
       `content` 字符串将不会被处理为消息段对象，此时字符串中的 cq 码将会直接生效，而不是被转义。

    .. admonition:: 警告
       :class: attention

       这个小技巧可以让你发送有效的 cq 码字符串。但是存在潜在的安全问题：
       如果将用户输入作为 cq 码字符串的一部分发送出去，这将会造成“注入攻击”！
       用户可以构造包含恶意图片、语音的 cq 码，让 bot 发送。

       任何时候启用 `cq_str` 选项，如需用到用户输入，务必校验。

    :param content: 发送内容
    :param cq_str: 是否以 cq 字符串发送（默认格式是消息段对象)
    :param overtime: 会话暂停的超时时间，超时将抛出 :class:`.SessionHupTimeout` 异常
    """
    await send(content, cq_str)
    try:
        await SESSION_LOCAL.hup(overtime)
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")


async def send_reply(
    content: str | CQMsgDict | list[CQMsgDict],
    cq_str: bool = False,
    wait: bool = False,
) -> Optional["ResponseEvent"]:
    """发送一条回复消息（在当前会话下自动定位发送目标）

    .. admonition:: 小技巧
       :class: note

       当启用了 `cq_str` 选项，且 `content` 参数为字符串时，
       `content` 字符串将不会被处理为消息段对象，此时字符串中的 cq 码将会直接生效，而不是被转义。

    .. admonition:: 警告
       :class: attention

       这个小技巧可以让你发送有效的 cq 码字符串。但是存在潜在的安全问题：
       如果将用户输入作为 cq 码字符串的一部分发送出去，这将会造成“注入攻击”！
       用户可以构造包含恶意图片、语音的 cq 码，让 bot 发送。

       任何时候启用 `cq_str` 选项，如需用到用户输入，务必校验。

    :param content: 发送内容
    :param cq_str: 是否以 cq 字符串发送（默认格式是消息段对象)
    :param wait: 是否等待发送后的响应
    :return:
       若指定等待 -> :class:`.ResponseEvent` 对象

       若未指定等待 -> :obj:`None`
    """
    try:
        content_arr = [reply_msg(SESSION_LOCAL.event.id)]
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")

    if isinstance(content, str):
        content_arr.append(text_msg(content))
    elif isinstance(content, dict):
        content_arr.append(content)
    else:
        content_arr.extend(content)
    return await send(content_arr, cq_str, wait)  # type: ignore


async def finish(
    content: str | CQMsgDict | list[CQMsgDict],
    cq_str: bool = False,
) -> None:
    """发送一条消息，然后直接结束当前事件处理方法（在当前会话下自动定位发送目标）

    只可在事件处理方法中使用。

    .. admonition:: 小技巧
       :class: note

       当启用了 `cq_str` 选项，且 `content` 参数为字符串时，
       `content` 字符串将不会被处理为消息段对象，此时字符串中的 cq 码将会直接生效，而不是被转义。

    .. admonition:: 警告
       :class: attention

       这个小技巧可以让你发送有效的 cq 码字符串。但是存在潜在的安全问题：
       如果将用户输入作为 cq 码字符串的一部分发送出去，这将会造成“注入攻击”！
       用户可以构造包含恶意图片、语音的 cq 码，让 bot 发送。

       任何时候启用 `cq_str` 选项，如需用到用户输入，务必校验。

    :param content: 发送内容
    :param cq_str: 是否以 cq 字符串发送（默认格式是消息段对象)
    """
    await send(content, cq_str)
    try:
        SESSION_LOCAL.destory()
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")
    raise DirectRetSignal("事件处理方法被安全地递归 return，请无视这个异常")


async def reply_finish(
    content: str | CQMsgDict | list[CQMsgDict],
    cq_str: bool = False,
) -> Optional["ResponseEvent"]:
    """发送一条回复消息，然后直接结束当前事件处理方法（在当前会话下自动定位发送目标）

    只可在事件处理方法中使用。

    .. admonition:: 小技巧
       :class: note

       当启用了 `cq_str` 选项，且 `content` 参数为字符串时，
       `content` 字符串将不会被处理为消息段对象，此时字符串中的 cq 码将会直接生效，而不是被转义。

    .. admonition:: 警告
       :class: attention

       这个小技巧可以让你发送有效的 cq 码字符串。但是存在潜在的安全问题：
       如果将用户输入作为 cq 码字符串的一部分发送出去，这将会造成“注入攻击”！
       用户可以构造包含恶意图片、语音的 cq 码，让 bot 发送。

       任何时候启用 `cq_str` 选项，如需用到用户输入，务必校验。

    :param content: 发送内容
    :param cq_str: 是否以 cq 字符串发送（默认格式是消息段对象)
    """
    try:
        content_arr = [reply_msg(SESSION_LOCAL.event.id)]
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")

    if isinstance(content, str):
        content_arr.append(text_msg(content))
    elif isinstance(content, dict):
        content_arr.append(content)
    else:
        content_arr.extend(content)
    await send(content_arr, cq_str)
    try:
        SESSION_LOCAL.destory()
    except LookupError:
        raise BotSessionError("当前 session 上下文不存在，因此无法使用本方法")
    raise DirectRetSignal("事件处理方法被安全地递归 return，请无视这个异常")



from .checker import (
    AtChecker,
    FriendReqChecker,
    GroupMsgLvlChecker,
    GroupReqChecker,
    MsgCheckerGen,
    MsgLvlChecker,
    NoticeTypeChecker,
    PrivateMsgLvlChecker,
)
from .formatter import ArgFormatter, FormatInfo
from .logger import BotLogger
from .matcher import ContainMatcher, EndMatcher, FullMatcher, RegexMatcher, StartMatcher
from .parser import CmdParser, CmdParserGen

__all__ = (
    "AtChecker",
    "FriendReqChecker",
    "GroupMsgLvlChecker",
    "GroupReqChecker",
    "MsgCheckerGen",
    "MsgLvlChecker",
    "NoticeTypeChecker",
    "PrivateMsgLvlChecker",
    "ArgFormatter",
    "FormatInfo",
    "BotLogger",
    "ContainMatcher",
    "EndMatcher",
    "FullMatcher",
    "RegexMatcher",
    "StartMatcher",
    "CmdParser",
    "CmdParserGen",
)
